/*
 * Copyright 2000 Hans Reiser
 */
#include "fsck.h"


//
// this will probably eventually replace pass4 in fsck
//


//
// FIXME: there is no way to know what hash function is used to order
// names in directory
//
#define MAX_GEN_NUMBER  127
#define SET_GENERATION_NUMBER(offset,gen_number) (GET_HASH_VALUE(offset)|(gen_number))
static __u32 get_third_component (char * name, int len)
{
    unsigned long res;

    if (!len || (len == 1 && name[0] == '.'))
	return DOT_OFFSET;
    if (len == 2 && name[0] == '.' && name[1] == '.')
	return DOT_DOT_OFFSET;

    res = keyed_hash (name, len);
    res = GET_HASH_VALUE(res);
    if (res == 0)
	res = 128;
    return res + MAX_GEN_NUMBER;
    
}


//
// looks for name in the directory dir, return 1 if name found, 0
// otherwise
//
static __u32 find_entry (struct key * dir, char * name)
{
    struct key entry_key;
//    struct path path;
    int retval;
    int i;

//    init_path (&path);
    INITIALIZE_PATH(path);

    entry_key.k_dir_id = dir->k_dir_id;
    entry_key.k_objectid = dir->k_objectid;
    set_offset (KEY_FORMAT_1, &entry_key, get_third_component (name, strlen (name)));
    set_type (KEY_FORMAT_1, &entry_key, TYPE_DIRENTRY);
    
    while (1) {
	struct buffer_head * bh;
	struct item_head * ih;
	struct reiserfs_de_head * deh;

	retval = usearch_by_key (&g_sb, &entry_key, &path);
	if (retval == ITEM_NOT_FOUND)
	    PATH_LAST_POSITION (&path) --;
	
	bh = PATH_PLAST_BUFFER (&path);
	ih = PATH_PITEM_HEAD (&path);
	deh = B_I_DEH (bh, ih);
	for (i = ih->u.ih_entry_count - 1; i >= 0; i --) {
	    if (strlen (name) != name_length (ih, &deh[i], i))
		continue;
	    if (!memcmp (name_in_entry (&deh[i], i), name, strlen (name))) {
		pathrelse (&path);
		return deh[i].deh_objectid;;
	    }
	}
	pathrelse (&path);
	if (get_offset (&ih->ih_key) == DOT_OFFSET)
	    return 0;
	set_offset (KEY_FORMAT_1, &entry_key, get_offset (&entry_key) - 1);
    }
    die ("find_entry: endless loop broken");
    return 0;
}
    

//
// add name pointing to 'key' to the directory 'dir'. FIXME: this will
// not work if there is a name in directory with the same value of
// hash function
//
static void add_entry (struct key * dir, char * name, struct key * key)
{
    struct key entry_key;
    char * entry;
    struct reiserfs_de_head * deh;
    int retval;
    INITIALIZE_PATH(path);
    int paste_size;

    // compose entry key to look for its place in the tree
    entry_key.k_dir_id = dir->k_dir_id;
    entry_key.k_objectid = dir->k_objectid;
    entry_key.u.k_offset_v1.k_offset = SET_GENERATION_NUMBER (get_third_component (name, strlen (name)), 0);
    entry_key.u.k_offset_v1.k_uniqueness = V1_DIRENTRY_UNIQUENESS;

    retval = usearch_by_entry_key (&g_sb, &entry_key, &path);
    if (retval == POSITION_FOUND) {
	printf ("add_entry: can not add name %s", name);
	pathrelse (&path);
	return;
    }

    entry = getmem (DEH_SIZE + ROUND_UP (strlen (name)));
    deh = (struct reiserfs_de_head *)entry;
    deh->deh_location = 0;
    deh->deh_offset = cpu_to_le32 (entry_key.u.k_offset_v1.k_offset);
    deh->deh_state = 0;
    mark_de_visible (deh);
    /* put key (ino analog) to de */
    deh->deh_dir_id = cpu_to_le32 (key->k_dir_id);
    deh->deh_objectid = cpu_to_le32 (key->k_objectid);
    memcpy ((char *)(deh + 1), name, strlen (name));
   
    if (SB_VERSION(&g_sb) == REISERFS_VERSION_2)
	paste_size = DEH_SIZE + ROUND_UP (strlen (name));
    else 
	paste_size = DEH_SIZE + strlen (name);
    reiserfsck_paste_into_item (&path, entry, paste_size);
}


/* mkreiserfs should have created this */
static __u32 make_lost_found_dir (void)
{
    int retval;
    INITIALIZE_PATH (path);
    struct key key;
    struct stat_data sd;
    struct item_head ih;
    char empty [EMPTY_DIR_SIZE];
    __u32 lost_found; // objectid of directory "/lost+found"
    int key_format;


     if (SB_VERSION(&g_sb) == REISERFS_VERSION_2)
	 key_format = KEY_FORMAT_2;
     else 
	 key_format = KEY_FORMAT_1;

    lost_found = find_entry (&root_key, "lost+found");
    if (lost_found)
	return lost_found;

    // key of the stat data
    key.k_dir_id = REISERFS_ROOT_OBJECTID;
    key.k_objectid = get_unused_objectid (&g_sb);
    if (!key.k_objectid)
	return 0;
    key.u.k_offset_v1.k_offset = 0;
    key.u.k_offset_v1.k_uniqueness = 0;
    
    /* compose item to insert: ih and stat data body */
    make_dir_stat_data (key_format, key.k_dir_id, key.k_objectid, &ih, (char *)&sd);
    
    retval = usearch_by_key (&g_sb, &key, &path);
    if (retval != ITEM_NOT_FOUND)
	return 0;
    
    reiserfsck_insert_item (&path, &ih, (char *)&sd);
    
    /* compose empty dir item and its item_head */
    set_offset (KEY_FORMAT_1, &ih.ih_key, DOT_OFFSET);
    set_type (KEY_FORMAT_1, &ih.ih_key, TYPE_DIRENTRY);
    ih.u.ih_entry_count = cpu_to_le16 (2);

    if (key_format == KEY_FORMAT_1) {
	ih.ih_item_len = cpu_to_le16 (EMPTY_DIR_SIZE_V1);
	make_empty_dir_item_v1 (empty, key.k_dir_id, key.k_objectid,
				REISERFS_ROOT_PARENT_OBJECTID, REISERFS_ROOT_OBJECTID);
    } else {
	ih.ih_item_len = cpu_to_le16 (EMPTY_DIR_SIZE);
	make_empty_dir_item (empty, key.k_dir_id, key.k_objectid,
			     REISERFS_ROOT_PARENT_OBJECTID, REISERFS_ROOT_OBJECTID);
    }

    // look for place to insert it
    set_offset (KEY_FORMAT_1, &key, DOT_OFFSET);
    set_type (KEY_FORMAT_1, &key, TYPE_DIRENTRY);
    retval = usearch_by_key (&g_sb, &key, &path);
    if (retval != ITEM_NOT_FOUND)
	die ("make_lost_found_dir: can not create empty dir body of \'lost+found\'");

    reiserfsck_insert_item (&path, &ih, empty);

    // add entry to the root directory
    add_entry (&root_key, "lost+found", &key);

    /* update root directory */
    if (usearch_by_key (&g_sb, &root_key, &path) != ITEM_FOUND)
	die ("make_lost_found_dir: can not find root directory");
    if (ih_key_format (get_ih (&path)) == KEY_FORMAT_1) {
	struct stat_data_v1 * root_sd = get_item (&path);

	root_sd->sd_size = cpu_to_le32 (le32_to_cpu (root_sd->sd_size) + DEH_SIZE + strlen ("lost+found"));
	root_sd->sd_nlink = cpu_to_le16 (le16_to_cpu (root_sd->sd_nlink) + 1);
    } else {
	struct stat_data * root_sd = get_item (&path);

	root_sd->sd_size = cpu_to_le64 (le64_to_cpu (root_sd->sd_size) + DEH_SIZE + strlen ("lost+found"));
	root_sd->sd_nlink = cpu_to_le32 (le32_to_cpu (root_sd->sd_nlink) + 1);
    }
    // FIXME: sd_blocks left unchanged
    mark_buffer_dirty (get_bh (&path), 1);
    pathrelse (&path);

    return key.k_objectid;
}



// key of '/lost+found' directory
struct key g_lost_found = {0, 0, {{0, 0}, }};


// FIXME: link only directories
static int link_lost (struct super_block * s, struct buffer_head ** path, int h)
{
    int i;
    struct item_head * ih;
    struct buffer_head * bh = path[h];
    char lost_name[80];

    ih = B_N_PITEM_HEAD (bh, 0);
    for (i = 0; i < B_NR_ITEMS (bh); i ++, ih++) {
	if (is_stat_data_ih (ih) && !is_item_reachable (ih) && S_ISDIR (B_I_STAT_DATA (bh, ih)->sd_mode)) {
	    struct key key = {0, 0, {{0, 0},}};

	    sprintf (lost_name, "%u_%u", le32_to_cpu (ih->ih_key.k_dir_id),
		     le32_to_cpu (ih->ih_key.k_objectid));
	    /* entry in lost+found directory will point to this key */
	    key.k_dir_id = ih->ih_key.k_dir_id;
	    key.k_objectid = ih->ih_key.k_objectid;
	    add_entry (&g_lost_found, lost_name, &key);
	    // FIXME: lost+found directory has incorrect sd_size and sd_nlink
	}
    }
    return 0;
}


void pass4 (struct super_block * s)
{
    struct buffer_head * path[MAX_HEIGHT] = {0,};
    int total[MAX_HEIGHT] = {0,};
    int cur[MAX_HEIGHT] = {0,};
    int h = 0;
    blocknr_t block = SB_ROOT_BLOCK (s);


    if (opt_lost_found == NO_LOST_FOUND)
	return ;

    if (!s->u.reiserfs_sb.s_hash_function) {
	s->u.reiserfs_sb.s_hash_function = keyed_hash;
	set_hash (s->u.reiserfs_sb.s_rs, TEA_HASH);
	reiserfs_warning ("pass4: do not know which hash is used to sort names, using default\n");
    }

    /* create lost+found directory (if it is not there) */
    g_lost_found.k_dir_id = root_key.k_objectid;
    g_lost_found.k_objectid = make_lost_found_dir ();
    if (!g_lost_found.k_objectid) {
	reiserfs_warning ("pass4: could not create lost+found directory\n");
	return;
    }

    printf ("Looking for lost files..");

    pass_through_tree (s, 0, link_lost);

    /* we have passed whole tree once again. Something have been added
       to the lost+found directory. Do semantic pass for it */
    printf ("Checking lost+found directory.."); fflush (stdout);
    check_semantic_tree (&g_lost_found, &root_key, 0);
    printf ("ok\n");
    
}






























