/*
 * Copyright 1996-1999 Hans Reiser
 */
#include "fsck.h"
#include <stdlib.h>


int step = 0; // 0 - find stat_data or any item ; 1 - find item ; 2 - already found


/* allocates buffer head and copy buffer content */
static struct buffer_head * make_buffer (int dev, int blocknr, int size, char * data)
{
  struct buffer_head * bh;

  bh = getblk (dev, blocknr, size);
  if (buffer_uptodate (bh))
    return bh;
//    die ("make_buffer: uptodate buffer found");
  memcpy (bh->b_data, data, size);
  set_bit (BH_Uptodate, (char *)&bh->b_state);
  return bh;
}

int find_not_of_one_file(struct key * to_find, struct key * key)
{
    if ((to_find->k_objectid != -1) &&
        (to_find->k_objectid != key->k_objectid))
        return 1;
    if ((to_find->k_dir_id != -1) &&
        (to_find->k_dir_id != key->k_dir_id))
        return 1;
    return 0;
}

static void find_a_key (struct key * key, __u64 offset, struct buffer_head * bh)
{
  int i;
  struct item_head * ih;
  struct reiserfs_de_head * deh;

  if (step == 2)
    return;

  for (i = 0; i < B_NR_ITEMS (bh); i ++)
  {
    ih = B_N_PITEM_HEAD (bh, i);
    if (find_not_of_one_file (key, &(ih->ih_key)))
      continue;

    if ( offset == MAX_OFFSET_V2 )
    {
        /* look for all items of this file */
        reiserfs_warning ("\nblock %d contains item of this file %k (item %d)\n", bh->b_blocknr, &(ih->ih_key), i);
            continue;
    }

    if (get_offset(&ih->ih_key) == offset)
    {
        reiserfs_warning ("\nblock %d contains key %k (item %d)\n", bh->b_blocknr, &(ih->ih_key), i);
        if ((key->k_objectid != -1) && (key->k_dir_id != -1))
        {
            /* key is specified precisely */
            step = 2;
            return;
        }
        continue;
    }

    if (is_direntry_ih (ih))
    {
      int j;

      for (j = 0, deh = B_I_DEH (bh, ih); j < ih_entry_count(ih); j ++, deh ++)
	if (deh->deh_offset == offset)
	{
	  reiserfs_warning ("\nblock %d contains key [%d %d %d 3] (item_pos %d, pos_in_item %d)\n",
	                                bh->b_blocknr, ih->ih_key.k_dir_id, ih->ih_key.k_objectid, offset, i, j);
          if ((key->k_objectid != -1) && (key->k_dir_id != -1))
          {
            /* key is specified precisely */
            step = 2;
  	    return;
  	  }
  	  continue;
	}
    }
  }
}

/*

static void find_a_key (struct key * key, __u64 offset, struct buffer_head * bh)
{
  int i;
  struct item_head * ih;
  struct reiserfs_de_head * deh;

  if (step == 2)
        return;

  for (i = 0; i < B_NR_ITEMS (bh); i ++) {
    ih = B_N_PITEM_HEAD (bh, i);
    if (not_of_one_file (key, &(ih->ih_key)))
      continue;

    if (!step)
    {
        if ( ((ih_version(ih) == ITEM_VERSION_2) && (offset == MAX_OFFSET_V2)) ||
             ((ih_version(ih) == ITEM_VERSION_1) && (offset == MAX_OFFSET_V1)) ) {
          // look for all items of this file
          reiserfs_warning ("\nblock %d contains item of this file %k (item %d)\n", bh->b_blocknr, key, i);
          return;
        }

        if (comp_le_keys (key, &(ih->ih_key)))
          continue;

        set_le_key_k_offset(ih_version(ih), key, offset);
        set_le_key_k_offset(ih_version(ih), key, TYPE_ANY);
        step = 1;
    }

    // key is specified precisely
    if (!comp_keys_3 (key, &(ih->ih_key), ih_version(ih))) {
      reiserfs_warning ("\nblock %d contains key %k (item %d)\n", bh->b_blocknr, key, i);
      step = 2;
      return;
    }
    if (is_direntry_le_key(ITEM_VERSION_1, key) && is_direntry_le_ih (ih)) {
      int j;
      
      deh = B_I_DEH (bh, ih);
      for (j = 0; j < I_ENTRY_COUNT (ih); j ++, deh ++)
	if (deh->deh_offset == le_key_k_offset(ITEM_VERSION_1, key)) {
	  reiserfs_warning ("\nblock %d contains key %k (item_pos %d, pos_in_item %d)\n", bh->b_blocknr, key, i, j);
          step = 2;
	  return;
	}
    }
  }
}
*/

/* analyse contents of indirect items. If it points to used blocks or
   to uninsertable node, which has to be inserted by items - we free
   those slots (putting 0-s), if not - mark pointed blocks as used */
static void handle_indirect_items (struct buffer_head * bh, int insert)
{
    int i, j;
    struct item_head * ih;

    for (i = 0, ih = B_N_PITEM_HEAD (bh, 0); i < B_NR_ITEMS (bh); i ++, ih ++) {
	if (is_indirect_ih(ih)) {
	    __u32 * unp;
	    
	    /* check each pointer to unformatted node, if it is in the tree already, put 0 here */
	    unp = (__u32 *)B_I_PITEM (bh, ih);
	    for (j = 0; j < ih_item_len(ih) / UNFM_P_SIZE; j ++)
	    {
		if (unp[j] == 0)
		    continue;
		if (unp[j] >= SB_BLOCK_COUNT (&g_sb) || /* invalid data block */
		    !was_block_used (unp[j]))	/* block is marked free in on
						   disk bitmap */
		{
		    unp[j] = 0;
		    mark_buffer_dirty (bh, 0);
		    continue;
		}
		
		if (is_block_formatted(unp[j]) ||
		    is_block_uninsertable(unp[j]))
		{
		    unp[j] = 0;
		    mark_buffer_dirty (bh, 0);
		    continue;
		}
		if (is_block_unformatted(unp[j]))
		{// duplicate content of unformatted node
		    struct buffer_head * bh_src, * bh_dest;
		    
		    if (!(bh_src = bread (g_sb.s_dev, unp[j], g_sb.s_blocksize)))
		    {
			unp[j] = 0;
			continue;
		    }

		    bh_dest = reiserfsck_get_new_buffer (bh->b_blocknr);
		    memcpy (bh_dest->b_data, bh_src->b_data, bh_src->b_size);

		    mark_block_unformatted (bh_dest->b_blocknr);
		    mark_buffer_dirty (bh_dest, 1);
		    mark_buffer_uptodate (bh_dest, 1);
		    
		    unp[j] = bh_dest->b_blocknr;
		    
		    brelse (bh_src);
		    brelse (bh_dest);

		    mark_buffer_dirty (bh, 0);
		    mark_buffer_uptodate (bh, 1);
		    continue;
		}
		/* ok, mark that block is in tree and that it is unformatted node */
		if (insert)
		{
		    mark_block_used (unp[j]);
		    mark_block_unformatted (unp[j]);
		}
	    }
	}
    }
}


int g_unaccessed_items = 0;

// these are 4 bits only
#define ITEM_REACHABLE 0
#define ITEM_UNREACHABLE 0xf

int is_item_reachable (struct item_head * ih)
{
    return (ih->ih_format.fsck_need == ITEM_REACHABLE) ? 1 : 0;
}


void mark_item_unreachable (struct item_head * ih)
{
    g_unaccessed_items ++;
    ih->ih_format.fsck_need = ITEM_UNREACHABLE;
//&&&&&&&&&&&&&&&&&&&&&&&&&&
    if (is_indirect_ih (ih))
      set_free_space (ih, 0); // 
//&&&&&&&&&&&&&&&&&&&&&&&&&&
    //  we do not need mark_buffer_dirty here because there is no way
    //  for this item to get into not dirtied node
}


void mark_item_reachable (struct item_head * ih, struct buffer_head * bh)
{
    g_unaccessed_items --;
    ih->ih_format.fsck_need = ITEM_REACHABLE;
    mark_buffer_dirty (bh, 0);
}


/* used when leaf is inserted into tree by pointer
   1. set sd_nlinks to 0 in all stat data items
   2. mark all items as unaccessed
   */
static void reset_nlinks (struct buffer_head * bh)
{ 
    int i;
    struct item_head * ih;

    ih = B_N_PITEM_HEAD (bh, 0);
    for (i = 0; i < B_NR_ITEMS (bh); i ++, ih ++) {
	mark_item_unreachable (ih);
	if (is_stat_data_ih (ih)) {
	    add_event (STAT_DATA_ITEMS);
        if (ih_key_format(ih) == KEY_FORMAT_1)
           ((struct stat_data_v1 *)B_I_PITEM(bh,ih))->sd_nlink = 0;
        else
            ((struct stat_data *)B_I_PITEM(bh,ih))->sd_nlink = 0;
	}
    }
    mark_buffer_dirty (bh, 0);
}


/* ubitmap.c */
extern int from_journal;

static void insert_pointer (struct buffer_head * bh, struct path * path)
{
  struct item_head * ih;
  char * body;
  int zeros_number;
  int retval;
  struct tree_balance tb;

  init_tb_struct (&tb, &g_sb, path, 0x7fff);
//  tb.preserve_mode = NOTHING_SPECIAL;

  /* fix_nodes & do_balance must work for internal nodes only */
  ih = 0;
//  retval = fix_nodes (0, M_INTERNAL, &tb, PATH_LAST_POSITION (path), ih);

  retval = fix_nodes (tb.transaction_handle, M_INTERNAL, &tb, ih);

  if (retval != CARRY_ON)
    die ("insert_pointer: no free space on device (retval == %d, used blocks from journal %d",
	 retval, from_journal);

  /* child_pos: we insert after position child_pos: this feature of the insert_child */
  /* there is special case: we insert pointer after
     (-1)-st key (before 0-th key) in the parent */
  if (PATH_LAST_POSITION (path) == 0 && path->pos_in_item == 0)
    PATH_H_B_ITEM_ORDER (path, 0) = -1;
  else {
    if (PATH_H_PPARENT (path, 0) == 0)
      PATH_H_B_ITEM_ORDER (path, 0) = 0;
/*    PATH_H_B_ITEM_ORDER (path, 0) = PATH_H_PPARENT (path, 0) ? PATH_H_B_ITEM_ORDER (path, 0) : 0;*/
  }

  ih = 0;
  body = (char *)bh;
  //memmode = 0;
  zeros_number = 0;

//  do_balance (0, &tb, path->pos_in_item, ih, body, M_INTERNAL, memmode, zeros_number);
  do_balance (0/*th*/, &tb, ih, body, M_INTERNAL, zeros_number);

  /* mark as used block itself and pointers to unformatted nodes */
  mark_block_used (bh->b_blocknr);

  /* this is for check only */
  mark_block_formatted (bh->b_blocknr);
  reset_nlinks (bh);

  handle_indirect_items (bh, 1);

  /* statistic */
  add_event (GOOD_LEAVES);
}


/* return 1 if left and right can be joined. 0 otherwise */
int balance_condition_fails (struct buffer_head * left, struct buffer_head * right)
{
  if (B_FREE_SPACE (left) >= B_CHILD_SIZE (right) -
      (are_items_mergeable (B_N_PITEM_HEAD (left, B_NR_ITEMS (left) - 1), B_N_PITEM_HEAD (right, 0), left->b_size) ? IH_SIZE : 0))
    return 1;
  return 0;
}


/* return 1 if new can be joined with last node on the path or with
   its right neighbor, 0 otherwise */
int balance_condition_2_fails (struct buffer_head * new, struct path * path)
{
  struct buffer_head * bh;
  struct key * right_dkey;
  int pos, used_space;

  bh = PATH_PLAST_BUFFER (path);


  if (balance_condition_fails (bh, new))
    /* new node can be joined with last buffer on the path */
    return 1;

  /* new node can not be joined with its left neighbor */

  right_dkey = uget_rkey (path);
  if (right_dkey == 0)
    /* there is no right neighbor */
    return 0;
  
  pos = PATH_H_POSITION (path, 1);
  if (pos == B_NR_ITEMS (bh = PATH_H_PBUFFER (path, 1))) {
    /* we have to read parent of right neighbor. For simplicity we
       call search_by_key, which will read right neighbor as well */
//    init_path (&path_to_right_neighbor);
    INITIALIZE_PATH(path_to_right_neighbor);

    if (usearch_by_key (&g_sb, right_dkey, &path_to_right_neighbor) != ITEM_FOUND)
      die ("get_right_neighbor_free_space: invalid right delimiting key");
    used_space =  B_CHILD_SIZE (PATH_PLAST_BUFFER (&path_to_right_neighbor));
    pathrelse (&path_to_right_neighbor);
  }
  else
    used_space = B_N_CHILD (bh, pos + 1)->dc_size;
  
  if (B_FREE_SPACE (new) >= used_space -
       (are_items_mergeable (B_N_PITEM_HEAD (new, B_NR_ITEMS (new) - 1), (struct item_head *)right_dkey, new->b_size) ? IH_SIZE : 0))
    return 1;

  return 0;
}

static void get_max_buffer_key (struct buffer_head * bh, struct key * key)
{
    struct item_head * ih;

    ih = B_N_PITEM_HEAD (bh, B_NR_ITEMS (bh) - 1);
    copy_key (key, &(ih->ih_key));

    if (is_direntry_key (key)) {
	/* copy 3-rd and 4-th key components of the last entry */
	//set_le_key_k_offset (ih_version(ih), key, B_I_DEH (bh, ih)[I_ENTRY_COUNT (ih) - 1].deh_offset);
	//set_le_key_k_type (ih_version(ih), key, TYPE_DIRENTRY);
	set_offset (KEY_FORMAT_1, key, 
			     le32_to_cpu (B_I_DEH (bh, ih)[ih_entry_count (ih) - 1].deh_offset));

    } else if (!is_stat_data_key (key))
	/* get key of the last byte, which is contained in the item */
	set_offset (key_format (key), key, get_offset (key) + get_bytes_number (ih, bh->b_size) - 1);
    //set_le_key_k_offset(ih_version(ih), key,
    //		    le_key_k_offset(ih_version(ih), key) + get_bytes_number (bh, ih, 0, CHECK_FREE_BYTES) - 1 );
}


/* inserts pointer to leaf into tree if possible. If not, marks node as uninsrtable */
static void try_to_insert_pointer_to_leaf (struct buffer_head * new_bh)
{
  struct path path;
  struct buffer_head * bh;			/* last path buffer */
  struct key * first_bh_key, last_bh_key;	/* first and last keys of new buffer */
  struct key last_path_buffer_last_key, * right_dkey;
  int ret_value;


  path.path_length = ILLEGAL_PATH_ELEMENT_OFFSET;
  if (is_block_used (new_bh->b_blocknr))
    /* block could get into tree already if its number was used by
       some indirect item */
    goto cannot_insert;

  first_bh_key = B_N_PKEY (new_bh, 0);

  /* try to find place in the tree for the first key of the coming node */
/*
  if (KEY_IS_DIRECTORY_KEY (first_bh_key))
    ret_value = search_by_entry_key (&g_sb, first_bh_key, &path);
  else if (KEY_IS_STAT_DATA_KEY (first_bh_key)) {
    ret_value = reiserfsck_search_by_key (&g_sb, first_bh_key, &path, comp_keys);
    path.pos_in_item = 0;
  } else
    ret_value = search_by_position (&g_sb, first_bh_key, &path);
  if (ret_value == KEY_FOUND || ret_value == ENTRY_FOUND || ret_value == BYTE_FOUND)
    goto cannot_insert;
*/
  ret_value = usearch_by_key (&g_sb, first_bh_key, &path);
  if (ret_value == ITEM_FOUND)
    goto cannot_insert;

  /* get max key in the new node */
  get_max_buffer_key (new_bh, &last_bh_key);

  bh = PATH_PLAST_BUFFER (&path);
  if (comp_keys (B_N_PKEY (bh, 0), &last_bh_key) == 1/* first is greater*/) {
    /* new buffer falls before the leftmost leaf */
    if (balance_condition_fails (new_bh, bh))
      goto cannot_insert;

    if (uget_lkey (&path) != 0 || PATH_LAST_POSITION (&path) != 0)
      die ("try_to_insert_pointer_to_leaf: bad search result");

    path.pos_in_item = 0;
    goto insert;
  }

  /* get max key of buffer, that is in tree */
  get_max_buffer_key (bh, &last_path_buffer_last_key);
  if (comp_keys (&last_path_buffer_last_key, first_bh_key) != -1/* second is greater */)
    /* first key of new buffer falls in the middle of node that is in tree */
    goto cannot_insert;

  right_dkey = uget_rkey (&path);
  if (right_dkey && comp_keys (right_dkey, &last_bh_key) != 1 /* first is greater */)
    goto cannot_insert;

  if (balance_condition_2_fails (new_bh, &path))
    goto cannot_insert;

insert:
  insert_pointer (new_bh, &path);
  goto out;

cannot_insert:
  /* statistic */
  add_event (UNINSERTABLE_LEAVES);
  handle_indirect_items (new_bh, 0);
  mark_block_uninsertable (new_bh->b_blocknr);

out:
  pathrelse (&path);
  brelse (new_bh);
  return;
}


static int tree_is_empty (void)
{
  return (SB_ROOT_BLOCK (&g_sb) == ~0) ? 1 : 0;
}


static void make_single_leaf_tree (struct buffer_head * bh)
{
    struct reiserfs_super_block * rs;
     
    rs = (struct reiserfs_super_block *)(SB_BUFFER_WITH_SB (&g_sb)->b_data);

    /* tree is empty, make tree root */
    set_root_block (rs, bh->b_blocknr);
    set_tree_height (rs, 2);

    mark_block_used (bh->b_blocknr);
    mark_buffer_dirty (SB_BUFFER_WITH_SB (&g_sb), 1);

    /* this is for check only */
    mark_block_formatted (bh->b_blocknr);
    
    /* set stat data nlinks fields to 0, mark all items as unaccessed,
       analyse contents of indirect items */
    reset_nlinks (bh);
    handle_indirect_items (bh, 1);
    
    /* statistic */
    add_event (GOOD_LEAVES);
    
    brelse (bh);
}

/*
static int is_internal_block_head1 (char *buf)
{
  int level;

  level = le16_to_cpu (((struct block_head *)buf)->blk_level);

  if (!(level > DISK_LEAF_NODE_LEVEL))
    return 0;
  if (!(level <= MAX_HEIGHT))
    return 0;
  return 1;
}
*/

/* reads the device by set of 8 blocks, takes leaves and tries to
   insert them into tree */
void build_the_tree (void)
{
    int i, j, k, max_k;
    struct buffer_head * bbh, * bh;
    __u32 handled_blocks = 0;
    struct si * saved_items = 0;
    char * data;

    if ( opt_fsck == 0 )
        fprintf (stderr, "Pass 1 - ");

        fflush(stdout);
    for (i = 0; i < SB_BMAP_NR (&g_sb); i ++)
	for (j = 0; j < g_sb.s_blocksize; j ++) {
        /* make sure, that we are not out of the device */
        if (i * g_sb.s_blocksize * 8 + j * 8 >= SB_BLOCK_COUNT (&g_sb))
            goto out_of_bitmap;
	    if (i * g_sb.s_blocksize * 8 + j * 8 + 8 > SB_BLOCK_COUNT (&g_sb))
            max_k = SB_BLOCK_COUNT (&g_sb) - (i * g_sb.s_blocksize * 8 + j * 8) - 1;
	    else
            max_k = 8;
        if (SB_AP_BITMAP (&g_sb)[i]->b_data[j] == 0) {
            /* all blocks are free */
           if (opt_what_to_scan == SCAN_USED_PART)
                continue;
	    }

	    if (opt_fsck_mode == FSCK_FAST_REBUILD &&
	        (SB_AP_BITMAP (&g_sb)[i]->b_data[j] & (~g_new_bitmap[i][j])) == 0)
            /* --fast-rebuild & !(any block was_used & not in tree yet)	 */
            continue;
	    	
	    bbh = bread (g_sb.s_dev, i * g_sb.s_blocksize + j, g_sb.s_blocksize * 8);
//	    printf("\n bread %d", (i * g_sb.s_blocksize + j) * 8 );
	    for (k = 0; k < max_k; k ++) {
            unsigned long block;
            int what_node;

            if ((SB_AP_BITMAP (&g_sb)[i]->b_data[j] & (1 << k)) == 0) {
                /* k-th block is free */
                if (opt_what_to_scan == SCAN_USED_PART)
                    continue;
            }
            block = i * g_sb.s_blocksize * 8 + j * 8 + k;
		
            if (opt_fsck_mode == FSCK_FAST_REBUILD &&
                (SB_AP_BITMAP (&g_sb)[i]->b_data[j] & (1 << k)) == 1 && is_block_used(block))
            /* this block is in tree already(we didn't remove it from the tree during check_internal_structure) */
            continue;

		    		
            if ( opt_fsck == 0 )
                print_how_far (&handled_blocks, g_blocks_to_read);
		
            if (not_data_block (&g_sb, block)) {
                /* skip not data area of the filesystem (journal,
                           bitmaps, reserved space) */
                continue;
            }

	
            if (bbh == 0)
            {
                bh = bread (g_sb.s_dev, block, g_sb.s_blocksize);
                if (bh == 0)
                {
                    mark_block_used(block);
                    continue;
                }
                data = bh->b_data;
            }else
                data = bbh->b_data + k * g_sb.s_blocksize;

            what_node = who_is_this (data, g_sb.s_blocksize);

            switch (what_node) {
            case THE_LEAF:
            case THE_INTERNAL:
                break;
            default:
                //
                continue;
            }
		
            if (is_internal_block_head(data) == 1)
              //if (is_internal_block_head1(bbh->b_data + k * g_sb.s_blocksize) == 1 && opt_fsck_mode == FSCK_REBUILD)
            {
                if (opt_fsck_mode == FSCK_REBUILD)
                {

//		if (b == 1 && opt_fsck_mode == FSCK_REBUILD) {
//  (struct block_head *)p_s_blkh->blk_level != DISK_LEAF_NODE_LEVEL)

/*		    g_found_internals ++;*/
        		    if (!is_block_used (block))
/*			g_freed_internals ++;*/
                    {
                        reiserfs_free_internal_block (&g_sb, block);
		            }
    		        else
			/* block is used in new tree already. There was an
			   indirect item, pointing to it. We keep information
			   about it for check only */
			/*mark_formatted_pointed_by_indirect (block)*/ ;
                }
    		    continue;
            }

            /* leaf node is selected */
            bh = make_buffer (g_sb.s_dev, block, g_sb.s_blocksize, data);
	    	/* */
            if (opt_fsck_mode == FSCK_FIND_ITEM) {
                find_a_key (&key_to_find, offset_to_find, bh);
                brelse (bh);
                continue;
    		}

            if (is_block_used (block)) {
		    /* block is used in new tree already. There was an indirect
		       item, pointing to it. We keep information about it for
		       check only */
		    /*	  mark_formatted_pointed_by_indirect (block);*/

                add_event (LEAVES_USED_BY_INDIRECT_ITEMS);
		    /* Rather than try to find UNP to this block we save its
		       items and will put them into tree at the end of pass 1 */
		
                for_all_items_in_node (save_item, &saved_items, bh);
        	    brelse (bh);
		        continue;
            }

            if (is_leaf_bad (bh)) {
                /* leaf is bad: directory item structure corrupted, or something else */
                /*	  mark_formatted_pointed_by_indirect (block);*/
		        if (opt_verbose)
            		reiserfs_warning ("\nbuild_the_tree: bad leaf encountered: %lu\n", bh->b_blocknr);
		        add_event (LEAVES_USED_BY_INDIRECT_ITEMS);
    		    /* Save good items only to put them into tree at the end of this pass */
		
	    	    for_all_items_in_node (save_item, &saved_items, bh);
		        brelse (bh);
		        continue;
    		}

            if (tree_is_empty () == 1) {
                make_single_leaf_tree (bh);
                continue;
            }

        	/* if the leaf node can not be inserted into tree by pointer,		
		       we postpone its insertion at the end of the pass 1 */
            try_to_insert_pointer_to_leaf (bh);
		
		/*&&&&&&&&&&&&&&&&*/
		//reiserfsck_check_pass1 ();
		/*&&&&&&&&&&&&&&&&*/
	    }

	    bforget (bbh);

	}


 out_of_bitmap:

/*
    check_bitmaps (&g_sb);

    printf ("bitmap scanning completed: allocated %d, freed %d *really freed %d, new internals %d, found internals %d (freed %d)\n",
	    g_allocated, g_freed, g_really_freed, g_new_internals, g_found_internals, g_freed_internals);fflush (stdout);
*/



    if (opt_fsck_mode == FSCK_FIND_ITEM)
	return;

    /* this checks what has been built (if -c option is set) */
/*    reiserfsck_check_pass1 ();*/

    /* put saved items into tree. These items were in leaves, those
       could not be inserted into tree because some indirect items point
       to those leaves. Rather than lookup for corresponding unfm
       pointers in the tree, we save items of those leaves and put them
       into tree separately */
    if ( opt_fsck == 0 )
	printf ("\nPass 1a - ");
	
    put_saved_items_into_tree (saved_items);

    if ( opt_fsck == 0 )
	printf ("done\n");

    /* end of pass 1 */
    if ( opt_fsck == 0 )
	printf ("\n");

    /* this works only if -c specified  */

    if (SB_ROOT_BLOCK(&g_sb) == -1)
        die("\n\n No one item was found, try --scan-whole-partition");

    reiserfsck_check_pass1 ();


    if (opt_stop_point == STOP_AFTER_PASS1)
	return;
	
    /* pass 2 */
    take_bad_blocks_put_into_tree ();

    reiserfsck_check_pass1 ();
}


/* pass the S+ tree of filesystem */
void recover_internal_tree (struct super_block * s)
{
    check_internal_structure(s);
    build_the_tree();
}



void rebuild_sb (int sb_not_found)
{
    int version;
    struct buffer_head * bh;
    struct reiserfs_super_block * rs;
    __u32 blocks;

    g_sb.s_blocksize = 4096;
    if(sb_not_found)
    {
        printf("\nwhat is version of ReiserFS you use[1-4]\n");
        printf("\t(1)   3.6.x\n");
        printf("\t(2) >=3.5.9\n");
        printf("\t(3) < 3.5.9 converted to new format\n");
        printf("\t(4) < 3.5.9\n");
        printf("\t(X)   exit\n");
	    if (scanf ("%d", &(version)) != 1)
    	    die ("rebuild_sb: specify a version through stdin");
        if (version < 1 || version > 4)
    	    die ("rebuild_sb: wrong version");
    	
        switch(version){
        case 1:
        case 2:
            bh = bread (g_sb.s_dev, (REISERFS_DISK_OFFSET_IN_BYTES / g_sb.s_blocksize), g_sb.s_blocksize);
            break;
        case 3:
        case 4:
            bh = bread (g_sb.s_dev, (2), g_sb.s_blocksize);
            break;
        default:
            exit(0);
        }
        if (!bh)
            die ("rebuild_sb: can't bread");
        rs = (struct reiserfs_super_block *)bh->b_data;
        g_sb.u.reiserfs_sb.s_rs = rs;
        SB_BUFFER_WITH_SB (&g_sb) = bh;
    }
    else
    {
        char * yes = 0;
        size_t n = 0;

        version = check_sb(&g_sb);
        printf("\nDo you want to remake your super block\n(say no if you use resizer)[Yes/no]:\n");
        if (getline (&yes, &n, stdin) != 4 || strcmp ("Yes\n", yes))
        	exit (0);
        rs = g_sb.u.reiserfs_sb.s_rs;
    }

    // set block number on the device and number of bitmap blocks needed to address all blocks
    blocks = (count_blocks ("", g_sb.s_blocksize, g_sb.s_dev) / 8) * 8;
    set_block_count (rs, blocks);
    //rs->s_block_count = cpu_to_le32(blocks);

    set_bmap_nr (rs, (blocks + (g_sb.s_blocksize * 8 - 1)) / (g_sb.s_blocksize * 8));
    set_journal_size(rs, JOURNAL_BLOCK_COUNT);

    //rs->s_bmap_nr = cpu_to_le16( blocks / (g_sb.s_blocksize * 8) +
    //      ((blocks % (g_sb.s_blocksize * 8)) ? 1 : 0) );

    switch (version){
    case 1:
	// super block v2 at 64k offset
	set_blocksize (rs, g_sb.s_blocksize);
	strncpy (rs->s_v1.s_magic, REISER2FS_SUPER_MAGIC_STRING,
		 strlen(REISER2FS_SUPER_MAGIC_STRING));
	set_journal_start (rs, get_journal_start_must (g_sb.s_blocksize));
	set_version (rs, REISERFS_VERSION_2);
	set_objectid_map_max_size (rs, (g_sb.s_blocksize - SB_SIZE) / sizeof(__u32) / 2 * 2);
	break;

    case 2:
	// super block v1 at 64k offset
	set_blocksize (rs, g_sb.s_blocksize);
	strncpy (rs->s_v1.s_magic, REISERFS_SUPER_MAGIC_STRING,
		 strlen(REISERFS_SUPER_MAGIC_STRING));
	set_journal_start (rs, get_journal_start_must (g_sb.s_blocksize));
	set_version (rs, REISERFS_VERSION_1);
	set_objectid_map_max_size (rs, (g_sb.s_blocksize - SB_SIZE_V1) / sizeof(__u32) / 2 * 2);
	break;

    case 3:
	// super block v2 at 8k offset
	set_blocksize (rs, g_sb.s_blocksize);
	strncpy (rs->s_v1.s_magic, REISER2FS_SUPER_MAGIC_STRING,
		 strlen(REISER2FS_SUPER_MAGIC_STRING));
	set_journal_start (rs, get_journal_old_start_must (rs));
	set_version (rs, REISERFS_VERSION_2);
	set_objectid_map_max_size (rs, (g_sb.s_blocksize - SB_SIZE) / sizeof(__u32) / 2 * 2);
	break;

    case 4:
	// super block v1 at 8k offset
	set_blocksize (rs, g_sb.s_blocksize);
	strncpy (rs->s_v1.s_magic, REISERFS_SUPER_MAGIC_STRING,
		 strlen(REISERFS_SUPER_MAGIC_STRING));
	set_journal_start (rs, get_journal_old_start_must (rs));
	set_version (rs, REISERFS_VERSION_1);
	set_objectid_map_max_size (rs, (g_sb.s_blocksize - SB_SIZE_V1) / sizeof(__u32) / 2 * 2);
	break;
    }

    mark_buffer_dirty(SB_BUFFER_WITH_SB (&g_sb), 1);
    reiserfs_warning("\nDo not forget to run reiserfsck --rebuild-tree\n\n");
}

/*
   check_sb and rebuild-sb don't touch these fields:
   __u32 s_journal_dev;
   __u32 s_journal_trans_max ;
   __u32 s_journal_block_count ;
   __u32 s_journal_max_batch ;
   __u32 s_journal_max_commit_age ;
   __u32 s_journal_max_trans_age ;

   others are checked and set in either rebuild_sb or rebuild-tree
*/


















