/*
 * Copyright 1996-1999 Hans Reiser
 */
#include "fsck.h"

//#include "reiserfs.h"

void for_all_items_in_node (action_on_item_t action, struct si ** si, struct buffer_head * bh)
{
  int i;
  struct item_head * ih;

  for (i = 0, ih = B_N_PITEM_HEAD (bh, 0); i < B_NR_ITEMS (bh); i ++, ih ++)
    action (si, ih, B_I_PITEM (bh,ih), bh->b_blocknr);

#if 0
  int j;

  for (i = B_NR_ITEMS (bh) / 2, j = i + 1; ; i --, j ++) {
    if (i >= 0) {
      ih = B_N_PITEM_HEAD (bh, i);
      action (si, ih, B_I_PITEM (bh,ih));
    }
    if (j < B_NR_ITEMS (bh)) {
      ih = B_N_PITEM_HEAD (bh, j);
      action (si, ih, B_I_PITEM (bh,ih));
    }

/*    check_buffer_queues ();*/

    if (i <= 0 && j >= B_NR_ITEMS (bh) - 1)
      break;
  }
#endif
}


/* insert sd item if it does not exist, overwrite it otherwise */
static void put_sd_item_into_tree (struct item_head * comingih, char * item, __u32 blocknr)
{
    struct item_head ih;
    struct path path;
//  struct buffer_head * path_bh;
//  int path_item_num;
    int key_format = ih_key_format (comingih);
    struct stat_data vsd;


    copy_key (&(ih.ih_key), &(comingih->ih_key));
    set_key_format(&ih, key_format);

    if (usearch_by_key (&g_sb, &(ih.ih_key), &path) == ITEM_FOUND) {

        if (ih_key_format (get_ih(&path)) != key_format)
        {
            // found stat data items are of different types
	    pathrelse (&path);
            return;
        }

	/* overwrite stat data in the tree */
	if (key_format == KEY_FORMAT_2)
	{
	    struct stat_data * psd = (struct stat_data *) get_item(&path);

	    if (psd->sd_nlink != 0)
		die ("put_sd_item_into_tree: all stat data in the tree (at this moment) "
		     "must have nllinks == 0 (not %d)", psd->sd_nlink);
	    if (psd->sd_mtime > ((struct stat_data *)item)->sd_mtime) {
		/* new sd is newer than the found one */
		memcpy (psd, item, SD_SIZE);
		psd->sd_nlink = 0;
	    }
	}else{
	    struct stat_data_v1 * psd = (struct stat_data_v1 *) get_item(&path);

	    if (psd->sd_nlink != 0)
		die ("put_sd_item_into_tree: all stat data in the tree (at this moment) must have nllinks == 0 (not %d)", psd->sd_nlink);
	    if (psd->sd_mtime > ((struct stat_data_v1 *)item)->sd_mtime) {
		/* new sd is newer than the found one */
		memcpy (psd, item, SD_V1_SIZE);
		psd->sd_nlink = 0;
		psd->sd_first_direct_byte = MAX_UL_INT;
	    }
	}
	mark_buffer_dirty (get_bh (&path), 0);
	pathrelse (&path);
    } else {
	/* insert stat data into tree */
	if (key_format == KEY_FORMAT_2)
	{
	    ih.ih_item_len = cpu_to_le16 (SD_SIZE);
	    memcpy (&vsd, item, SD_SIZE);
	    vsd.sd_nlink = 0;
	}
	else
	{
	    struct stat_data_v1 * sd = (struct stat_data_v1 *)&vsd;

	    ih.ih_item_len = cpu_to_le16 (SD_V1_SIZE);
	    memcpy (sd, item, SD_V1_SIZE);
	    sd->sd_nlink = 0;
	    sd->sd_first_direct_byte = MAX_UL_INT;
	}

	set_key_format (&ih, key_format);
	set_free_space(&ih, MAX_US_INT);
	mark_item_unreachable (&ih);

	reiserfsck_insert_item (&path, &ih, (const char *)&vsd);
	add_event (STAT_DATA_ITEMS);
    }
}


/* Keyed 32-bit hash function using TEA in a Davis-Meyer function */
/*
static unsigned long get_third_component (char * name, int len)
{
  if (!len || (len == 1 && name[0] == '.'))
    return DOT_OFFSET;
  if (len == 2 && name[0] == '.' && name[1] == '.')
    return DOT_DOT_OFFSET;
  return keyed_hash (name, len);
}
*/

static int reiserfsck_find_entry (struct key * key, struct reiserfs_de_head * deh,
				  struct path * path)
{
    struct key entry_key;

    copy_key (&entry_key, key);
    set_offset (KEY_FORMAT_1, &entry_key, deh->deh_offset);
    set_type (KEY_FORMAT_1, &entry_key, TYPE_DIRENTRY);

    return usearch_by_entry_key (&g_sb, &entry_key, path);
}


/* this tries to put each item entry to the tree, if there is no items of
   the directory, insert item containing 1 entry */
static void put_directory_item_into_tree (struct item_head * comingih, char * item)
{
    /*  struct item_head * ih;*/
    struct reiserfs_de_head * deh;
    int i, retval;
    struct path path;
    int size;
    char * buf, * entry;
    struct item_head tmpih;

    /*ih = B_N_PITEM_HEAD (bh, item_num);*/
    deh = (struct reiserfs_de_head *)item;/*B_I_DEH (bh, comingih);*/

    for (i = 0; i < ih_entry_count (comingih); i ++, deh ++) {
	entry = item + deh->deh_location;

	{
	    // make sure we are looking for the name in proper place
	    char * name = name_in_entry (deh, i);
	    int namelen = name_length (comingih, deh, i);

	    if (namelen == 1 && name[0] == '.') {
		if (deh_offset (deh) != DOT_OFFSET)
		    deh->deh_offset = cpu_to_le32 (DOT_OFFSET);
	    } else if (namelen == 2 && name[0] == '.' && name[1] == '.') {
		if (deh_offset (deh) != DOT_DOT_OFFSET)
		    deh->deh_offset = cpu_to_le32 (DOT_DOT_OFFSET);
	    } else if (!good_name (hashes [g_real_hash], name, namelen, deh_offset (deh))) {
		// FIXME: something wrong might be here with collisions
		reiserfs_warning ("put_directory_item_into_tree: "
				  "name %s does not match to its offset, fixed\n", bad_name (name, namelen));
		deh->deh_offset = cpu_to_le32 (hashes [g_real_hash] (name, namelen));
	    }
	}

	retval = reiserfsck_find_entry (&(comingih->ih_key), deh, &path);


	switch (retval) {
	case POSITION_FOUND:
	    pathrelse (&path);
	    break;

	case POSITION_NOT_FOUND:
	    /* paste_into_item accepts entry to paste as buffer, beginning
	       with entry header and body, that follows it */
	    buf = getmem (size = entry_length (comingih, deh, i) + DEH_SIZE);
	    memcpy (buf, deh, DEH_SIZE);
	    ((struct reiserfs_de_head *)buf)->deh_location = 0;
	    memcpy (buf + DEH_SIZE, entry, size - DEH_SIZE);

	    reiserfsck_paste_into_item (&path, buf, size);

	    freemem (buf);
	    break;

	case DIRECTORY_NOT_FOUND:
	    // insert item containing on entry into the tree
	    buf = getmem (size = entry_length(comingih, deh, i) + DEH_SIZE);
	    memcpy (buf, deh, DEH_SIZE);
	    ((struct reiserfs_de_head *)buf)->deh_location = cpu_to_le16 (DEH_SIZE);
	    memcpy (buf + DEH_SIZE, entry, size - DEH_SIZE);

	    set_key_format (&tmpih, KEY_FORMAT_1);
	    copy_key (&(tmpih.ih_key), &(comingih->ih_key));
	    tmpih.ih_item_len = cpu_to_le16 (size);
	    tmpih.u.ih_entry_count = cpu_to_le16 (1);
	    mark_item_unreachable (&tmpih);
      
	    reiserfsck_insert_item (&path, &tmpih, buf);

	    freemem (buf);
	    break;

	case REGULAR_FILE_FOUND:
	    /* this should never happen. */
	    goto end;
	}

	/*&&&&&&&&&&&&&&&&&*/
/*    reiserfsck_check_pass1 ();*/
	/*&&&&&&&&&&&&&&&&&*/
    }
 end:

}

#if 0
/* If item is item of regular file (direct or indirect item) - this
   file is in tree (with first byte) - write to it. If this file is in
   tree (without first byte) - delete what we have in tree, create
   file again keeping what we already had in tree this file is not in
   tree - create hole at the beginning of file if necessary and write
   to file */
void put_regular_file_item_into_tree (struct item_head * ih, char * item, __u32 blocknr)
{
  reiserfsck_file_write (ih, item, blocknr);
}
#endif

void insert_item_separately (struct si ** si, struct item_head * ih, char * item, __u32 blocknr)
{
    // FIXME: refuse transparently bad items
    if (ih->ih_key.k_dir_id == ih->ih_key.k_objectid)
	return;
    if (is_stat_data_ih (ih)) {
	put_sd_item_into_tree (ih, item, blocknr);
    } else if (is_direntry_ih (ih)) {
	put_directory_item_into_tree (ih, item);
    } else {
	reiserfsck_file_write/*put_regular_file_item_into_tree*/ (ih, item, blocknr);
    }
}


/* uninsertable blocks are marked by 0s in
   g_uninsertable_leaf_bitmap during the pass 1. They still must be not in the tree */
void take_bad_blocks_put_into_tree (void)
{
    struct buffer_head * bh;
    int i, j;
    __u32 bb_counter = 0;
    int what_node;

    if ( opt_fsck == 0 )
	fprintf (stderr, "Pass 2 - ");


    for (i = 0; i < SB_BMAP_NR (&g_sb); i ++) {
	j = find_first_zero_bit (g_uninsertable_leaf_bitmap[i], g_sb.s_blocksize * 8);
	while (j < g_sb.s_blocksize * 8) {
	    bh = bread (g_sb.s_dev, i * g_sb.s_blocksize * 8 + j, g_sb.s_blocksize);
	    if (bh == 0)
                reiserfs_panic (&g_sb, "take_bad_blocks_put_into_tree: unable to read %lu block on device 0x%x\n",
			      i * g_sb.s_blocksize * 8 + j, g_sb.s_dev);

	    if (is_block_used (bh->b_blocknr)) {
		reiserfs_warning ("take_bad_blocks_put_into_tree: block %d can not be in tree\n", bh->b_blocknr);
		goto next;
	    }
	    /* this must be leaf */
	    what_node = who_is_this (bh->b_data, g_sb.s_blocksize);
	    if (what_node != THE_LEAF) { // || B_IS_KEYS_LEVEL(bh)) {
		reiserfs_warning ("take_bad_blocks_put_into_tree: buffer (%b %z) must contain leaf\n", bh, bh);
		goto next;
	    }
	
	    for_all_items_in_node (insert_item_separately, 0, bh);

	    if ( opt_fsck == 0 )
		print_how_far (&bb_counter, get_event (UNINSERTABLE_LEAVES));
	next:
	    brelse (bh);

	    j = find_next_zero_bit (g_uninsertable_leaf_bitmap[i], g_sb.s_blocksize * 8, j + 1);
	}
    }

    if (bb_counter != get_event (UNINSERTABLE_LEAVES))
	die ("take_bad_blocks_put_into_tree: found bad block %d, must be %d", 
	     bb_counter, get_event (UNINSERTABLE_LEAVES));

    if ( opt_fsck == 0 )
	printf ("\n");
}




















































