/*
 * Copyright 1996, 1997, 1998 Hans Reiser
 */
#include "fsck.h"


static void get_next_key (struct path * path, int i, struct key * key)
{
    struct buffer_head * bh = PATH_PLAST_BUFFER (path);
    struct key maxkey;
    struct key * rkey;

    memset (&maxkey, 0xff, KEY_SIZE);
    if (i < B_NR_ITEMS (bh) - 1) {
	copy_key (key, B_N_PKEY (bh, i + 1));
	return;
    }

    rkey = uget_rkey (path);
    if (rkey) {
	copy_key (key, rkey);
	/*
	if (comp_keys (key, B_PRIGHT_DELIM_KEY (bh)) != KEYS_IDENTICAL) {
	    add_event (FIXED_RIGHT_DELIM_KEY);
	    copy_key (B_PRIGHT_DELIM_KEY (bh), key);
	    mark_buffer_dirty (bh, 0);
	}
	*/
    } else {
	/*
	if (comp_le_keys (&maxkey, B_PRIGHT_DELIM_KEY (bh)) != KEYS_IDENTICAL) {
	    printf ("get_next_key: Hmm, max key not found in the tree\n");
	    copy_key (B_PRIGHT_DELIM_KEY (bh), &maxkey);
	    mark_buffer_dirty (bh, 0);
	}
	*/
	copy_key (key, &maxkey);
    }
}

/* this is pass 4 */
int check_unaccessed_items (void)
{
    struct key key;
    struct path path;
    int i;
    struct buffer_head * bh;
    struct item_head * ih;
    __u32 passed = 0;
    __u32 * p;
    int length = REISERFS_FULL_KEY_LEN;

    if (opt_stop_point != STOP_DEFAULT)
	return 0;

    path.path_length = ILLEGAL_PATH_ELEMENT_OFFSET;
    copy_key (&key, &g_root_directory_key);

    if ( opt_fsck == 0 )
	fprintf (stderr, "Pass 4 - ");

    while (usearch_by_key (&g_sb, &key, &path) == ITEM_FOUND) {
	bh = PATH_PLAST_BUFFER (&path);
	for (i = PATH_LAST_POSITION (&path), ih = PATH_PITEM_HEAD (&path); i < B_NR_ITEMS (bh); i ++, ih ++) {
	    if (!is_item_reachable (ih)) {

		get_next_key (&path, i, &key);

		add_event (UNACCESSED_ITEMS);
		if (is_stat_data_ih (ih))
		    g_fsck_info.fs_stat_data_items --;
	
		PATH_LAST_POSITION (&path) = i;
		reiserfsck_delete_item (&path);

		goto cont;
	    }
	    if ((is_stat_data_ih (ih)) && opt_fsck == 0) {
		print_how_far (&passed, get_event (STAT_DATA_ITEMS));
	    }
	}
	get_next_key (&path, i - 1, &key);
	pathrelse (&path);

/*fu_check ();*/

    cont:
    }

    p = (__u32 *)&key;

    for( ; length--; ++p )
    {
      if ( *p != (__u32)~0 )
	reiserfs_panic (0, "check_unaccessed_items: invalid exit key %k", &key);	
    }
/*
    if (key.k_dir_id != MAX_UL_INT || key.k_objectid != MAX_UL_INT ||
	key.k_offset != MAX_UL_INT || key.k_uniqueness != MAX_UL_INT) {
	reiserfs_panic (0, "check_unaccessed_items: invalid exit key %k", &key);
    }
*/

    pathrelse (&path);

    if ( opt_fsck == 0 )
	printf ("\n");

    return 0;
}










