/*
 * Copyright 1996-1999 Hans Reiser
 */
#include "fsck.h"
#include <time.h>
#include <math.h>

struct path_key
{
    struct short_key
    {
        __u32 k_dir_id;
        __u32 k_objectid;
    } key;
    struct path_key * next, * prev;
};
struct path_key * head_key = NULL;
struct path_key * tail_key = NULL;

void check_path_key(struct key * key)
{
    struct path_key * cur = head_key;

    while(cur != NULL)
    {
        if (!comp_short_keys(&cur->key, key))
            die("\nsemantic check: loop found %k", key);
        cur = cur->next;
    }
}

void add_path_key(struct key * key)
{
    check_path_key(key);

    if (tail_key == NULL)
    {
        tail_key = getmem(sizeof(struct path_key));
        head_key = tail_key;
        tail_key->prev = NULL;
    }else{
        tail_key->next = getmem(sizeof(struct path_key));
        tail_key->next->prev = tail_key;
        tail_key = tail_key->next;
    }
    copy_short_key (&tail_key->key, key);
    tail_key->next = NULL;
}

void del_path_key()
{
    if (tail_key == NULL)
        die("wrong path_key structure");

    if (tail_key->prev == NULL)
    {
        freemem(tail_key);
        tail_key = head_key = NULL;
    }else{
        tail_key = tail_key->prev;
        freemem(tail_key->next);
        tail_key->next = NULL;
    }
}

static void print (char * dir_name, __u32 len)
{
    int i;

    printf("/");
    for (i = 0; i<len; i++, dir_name++)
        printf ("%c", *dir_name);
    fflush (stdout);
}

static void erase_print (__u32 len)
{
    int i;

    for (i = 0; i<=len; i++)
        printf("\b");
    for (i = 0; i<=len; i++)
        printf(" ");
    for (i = 0; i<=len; i++)
        printf("\b");
    fflush (stdout);
}

/* path is path to stat data */
static void check_regular_file (struct path * path, void * sd)
{
    int mark_passed_items;
    struct key key, sd_key;
    __u64 size, min_size, saved_size;
    __u32 blocks, saved_blocks;
    struct buffer_head * bh = PATH_PLAST_BUFFER (path);/* contains stat data */
    struct item_head * ih = PATH_PITEM_HEAD (path);/* stat data item */
    int temp_res;
    int symlnk = 0;

    if (ih_key_format (ih) == KEY_FORMAT_2)
    {
	struct stat_data * sd_v2 = sd;
	
	if (sd_v2->sd_nlink == 0) {
	    /*    print_how_far (&stat_datas, get_event (STAT_DATA_ITEMS));*/
	    if (S_ISREG (sd_v2->sd_mode))
		add_event (REGULAR_FILES);
	    else if (S_ISLNK (sd_v2->sd_mode))
	    {
		symlnk = 1;
		add_event (SYMLINKS);
	    }
	    else
		add_event (OTHERS);

	    saved_size = sd_v2->sd_size;
	    saved_blocks = sd_v2->sd_blocks;
	
/*            if (mark_id_as_used(ih->ih_key.k_objectid))
                reiserfs_warning("check_regular_file: objectid is used already%k \n", &ih->ih_key);*/
	
	    if (opt_fsck_mode != FSCK_SEMANTIC)
	    {
                sd_v2->sd_nlink ++;
	        mark_item_reachable (ih, bh);
	        mark_objectid_as_used (&g_sb, ih->ih_key.k_objectid);
            }else{
                if (!is_objectid_used (&g_sb, ih->ih_key.k_objectid))
                     reiserfs_panic(&g_sb, "check_regular_file: unused objectid found %k\n", &ih->ih_key);
            }


	    copy_key (&key, &(ih->ih_key));
//	    copy_key (&sd_key, &key);

	    pathrelse (path);
	    if (are_file_items_correct (&key, KEY_FORMAT_2, &size, &min_size, &blocks,
	                mark_passed_items = ((opt_fsck_mode == FSCK_SEMANTIC)?0:1), symlnk, sd_v2->sd_size) != 1) {
		/* unpassed items will be deleted in pass 4 as they left unaccessed */
		add_event (INCORRECT_REGULAR_FILES);
	    }
	    /* are_file_items_correct could perform indirect_to_direct, bh could be changed */

	
	    copy_key (&sd_key, &key);
	    //set_le_key_k_offset (ih_version(ih), &sd_key, SD_OFFSET);
	    //set_le_key_k_type (ih_version(ih), &sd_key, TYPE_STAT_DATA);

	    // FIXME: we should look for stat data again only if we
	    // have to update file size
	    //
	    if ((size != 0) && ((saved_size < min_size) || (blocks != saved_blocks)))
	    {
	        if (opt_fsck_mode == FSCK_SEMANTIC)
	        {
	            if (saved_size < min_size)
	                reiserfs_panic(&g_sb, "check_regular_file: %k has bad size\n", &sd_key);
	            else
	                reiserfs_panic(&g_sb, "check_regular_file: %k has bad block number\n", &sd_key);
	        }
	
                set_type_and_offset (KEY_FORMAT_2, &sd_key, SD_OFFSET, TYPE_STAT_DATA);
	
		if (usearch_by_key (&g_sb, &(sd_key), path) != ITEM_FOUND)
		    die ("check_regular_file: stat data not found");
		
		bh = PATH_PLAST_BUFFER (path);
		ih = PATH_PITEM_HEAD (path);
		sd_v2 = B_N_STAT_DATA (bh, PATH_LAST_POSITION (path));
	    }
	    else
		sd_v2 = 0;

	    /* set correct size */
	
//	        printf(" size: was %d;", le64_to_cpu (sd_v2->sd_size));
//	        printf("will %d;", size);
	
	    if ((sd_v2 !=0) && (size != 0) && (le64_to_cpu (sd_v2->sd_size) < min_size))
	    {
//	        printf(" Fixed on %d",  size - le64_to_cpu (sd_v2->sd_size));
		add_event (FIXED_SIZE_FILES);
		sd_v2->sd_size = cpu_to_le64 (size);
		mark_buffer_dirty (bh, 0);
	    }
	    if ((sd_v2 !=0) && (blocks != le32_to_cpu (sd_v2->sd_blocks)))
	    {
	        sd_v2->sd_blocks = cpu_to_le32 (blocks);
	    }
	
	} else if (opt_fsck_mode != FSCK_SEMANTIC){
	    if ((temp_res = is_item_reachable (ih)) == 0)
		die ("check_regular_file: stat data item must be accessed");
	    sd_v2->sd_nlink ++;
	    mark_buffer_dirty (bh, 0);
	}
    }else{
	struct stat_data_v1 * sd_v1 = sd;

	if (sd_v1->sd_nlink == 0) {
	    /*    print_how_far (&stat_datas, get_event (STAT_DATA_ITEMS));*/
	    if ((sd_v1->sd_mode & S_IFMT) == S_IFREG)
		add_event (REGULAR_FILES);
	    else if ((sd_v1->sd_mode & S_IFMT) == S_IFLNK)
	    {
		add_event (SYMLINKS);
		symlnk = 1;
	    }
	    else
		add_event (OTHERS);
		
/*            if (mark_id_as_used(ih->ih_key.k_objectid))
                reiserfs_warning("check_regular_file: objectid is used already%k \n", &ih->ih_key);*/
	
	    if (opt_fsck_mode != FSCK_SEMANTIC)
	    {
	        sd_v1->sd_nlink ++;
	        mark_item_reachable (ih, bh);
	        mark_objectid_as_used (&g_sb, ih->ih_key.k_objectid);
            }else{
                if (!is_objectid_used (&g_sb, ih->ih_key.k_objectid))
                     reiserfs_panic(&g_sb, "check_regular_file: unused objectid found %k\n", &ih->ih_key);
            }

	    copy_key (&key, &(ih->ih_key));

	    pathrelse (path);
	    if (are_file_items_correct (&key, ih_key_format (ih), &size, &min_size, &blocks,
	                        mark_passed_items = ((opt_fsck_mode == FSCK_SEMANTIC)?0:1), symlnk, sd_v1->sd_size) != 1) {
		/* unpassed items will be deleted in pass 4 as they left unaccessed */
		add_event (INCORRECT_REGULAR_FILES);
	    }
	    /* are_file_items_correct could perform indirect_to_direct, bh could be changed */

	    //copy_key (&sd_key, &key);
	    //set_le_key_k_offset (ih_version(ih), &sd_key, SD_OFFSET);
	    //set_le_key_k_type (ih_version(ih), &sd_key, SD_UNIQUENESS);
	    copy_key (&sd_key, &key);
	    set_type_and_offset (KEY_FORMAT_1, &sd_key, SD_OFFSET, TYPE_STAT_DATA);

	    if (usearch_by_key (&g_sb, &sd_key, path) != ITEM_FOUND)
		die ("check_regular_file: stat data not found");
	
	    bh = PATH_PLAST_BUFFER (path);
	    ih = PATH_PITEM_HEAD (path);
	    sd_v1 = (struct stat_data_v1 *)B_N_STAT_DATA (bh, PATH_LAST_POSITION (path));

	    /* set correct size */
	    if (le32_to_cpu (sd_v1->sd_size) < min_size) {
//	        printf(" Fixed on %d",  size - le64_to_cpu (sd_v1->sd_size));
	        if (opt_fsck_mode == FSCK_SEMANTIC)
                    reiserfs_panic(&g_sb, "check_regular_file: %k has bad size\n", &sd_key);

		add_event (FIXED_SIZE_FILES);
		sd_v1->sd_size = cpu_to_le32 (size);
		mark_buffer_dirty (bh, 0);
	    }
	    if (blocks != le32_to_cpu (sd_v1->u.sd_blocks))
	    {
	        if (opt_fsck_mode == FSCK_SEMANTIC)
                    reiserfs_panic(&g_sb, "check_regular_file: %k has bad block number\n", &sd_key);
	
	        sd_v1->u.sd_blocks = cpu_to_le32 (blocks);
	    }
	    /* set first direct byte field of stat data (if it is set incorrect) */
	    if (size == 0 || is_indirect_key (&key)) {
		/* there are no direct items in file */
		
		if (sd_v1->sd_first_direct_byte != MAX_UL_INT)
		{
	            if (opt_fsck_mode == FSCK_SEMANTIC)
	                reiserfs_panic(&g_sb, "check_regular_file: %k has bad sd_first_direct_byte\n", &sd_key);
	
		    sd_v1->sd_first_direct_byte = MAX_UL_INT;
		    mark_buffer_dirty (bh, 0);
		}
	    } else {
		/* there is at least one direct item */
		if (sd_v1->sd_first_direct_byte != 
		    get_offset (&key) - (get_offset (&key) % g_sb.s_blocksize - 1))
		{
	            if (opt_fsck_mode == FSCK_SEMANTIC)
	                reiserfs_panic(&g_sb, "check_regular_file: %k has bad sd_first_direct_byte\n", &sd_key);
	
		    sd_v1->sd_first_direct_byte = cpu_to_le32 (get_offset (&key) - (get_offset (&key) % g_sb.s_blocksize - 1));
		    mark_buffer_dirty (bh, 0);
		}
	    }
	} else if (opt_fsck_mode != FSCK_SEMANTIC){
	    if (!is_item_reachable (ih))
		die ("check_regular_file: stat data item must be accessed");
	    sd_v1->sd_nlink ++;
	    mark_buffer_dirty (bh, 0);
	}
    }
}


static int is_rootdir_key (struct key * key)
{
    if (comp_keys (key, &root_key))
	return 0;
    return 1;
}

static int is_rootdir_entry_key (struct key * key)
{
    if (comp_short_keys (key, &root_key))
	return 0;
    return 1;
}


/* when root direcotry can not be found */
static void create_root_directory (struct path * path)
{
    struct item_head ih;
    void * sd;
    struct stat_data sd_v2;

    sd = &sd_v2;

    /* insert stat data item */
    copy_key (&ih.ih_key, &root_key);
    set_free_space(&ih,MAX_US_INT);
    mark_item_unreachable (&ih);

    if (SB_VERSION(&g_sb) == REISERFS_VERSION_2)
    {
	ih.ih_item_len = cpu_to_le16 (SD_SIZE);
	set_key_format (&ih, KEY_FORMAT_2);

	sd_v2.sd_mode = cpu_to_le16 (S_IFDIR + 0755);
	sd_v2.sd_nlink = 0;
	sd_v2.sd_uid = 0;
	sd_v2.sd_gid = 0;
	sd_v2.sd_size = cpu_to_le64 (EMPTY_DIR_SIZE);
	sd_v2.sd_atime = sd_v2.sd_ctime = sd_v2.sd_mtime = cpu_to_le32 (time (NULL));
	sd_v2.sd_blocks = 0;
	sd_v2.u.sd_rdev = 0;
    }
    else
    {
	struct stat_data_v1 * sd_v1;

	sd_v1 = sd;
	ih.ih_item_len = cpu_to_le16 (SD_V1_SIZE);
	set_key_format (&ih, KEY_FORMAT_1);

	sd_v1->sd_mode = cpu_to_le16 (S_IFDIR + 0755);
	sd_v1->sd_nlink = 0;
	sd_v1->sd_uid = 0;
	sd_v1->sd_gid = 0;
	sd_v1->sd_size = cpu_to_le32 (EMPTY_DIR_SIZE_V1);
	sd_v1->sd_atime = sd_v1->sd_ctime = sd_v1->sd_mtime = cpu_to_le32 (time (NULL));
	sd_v1->u.sd_blocks = 0;
	sd_v1->u.sd_rdev = 0;
	sd_v1->sd_first_direct_byte = MAX_UL_INT;

    }
    reiserfsck_insert_item (path, &ih, sd);
}


static void paste_dot_and_dot_dot (struct path * path)
{
    char dir[EMPTY_DIR_SIZE];
    struct reiserfs_de_head * deh;
    struct key key;
  
    copy_key (&key, &root_key);

    deh = (struct reiserfs_de_head *)dir;
    deh[0].deh_offset = DOT_OFFSET;
    deh[0].deh_dir_id = root_key.k_dir_id;
    deh[0].deh_objectid = root_key.k_objectid;
    deh[0].deh_state = 0;
    set_bit (DEH_Visible, &(deh[0].deh_state));
    dir[DEH_SIZE] = '.';

    if (SB_VERSION(&g_sb) == REISERFS_VERSION_2)
    {
      reiserfsck_paste_into_item (path, dir, DEH_SIZE + ROUND_UP (strlen (".")));
    }else{
      reiserfsck_paste_into_item (path, dir, DEH_SIZE + 1);
    }

    set_type_and_offset (KEY_FORMAT_1, &key, DOT_DOT_OFFSET, TYPE_DIRENTRY);
    //set_le_key_k_offset(ITEM_VERSION_1, &key, DOT_DOT_OFFSET);
    //set_le_key_k_type(ITEM_VERSION_1, &key, TYPE_DIRENTRY);

    if (usearch_by_entry_key (&g_sb, &key, path) == POSITION_FOUND) {
	reiserfs_warning ("paste_dot_and_dot_dot: \"..\" found\n");
	pathrelse (path);
	return;
    }
    deh[0].deh_offset = DOT_DOT_OFFSET;
    deh[0].deh_dir_id = 0;
    deh[0].deh_objectid = root_key.k_dir_id;
    deh[0].deh_state = 0;
    set_bit (DEH_Visible, &(deh[0].deh_state));
    dir[DEH_SIZE] = '.';
    dir[DEH_SIZE + 1] = '.';

    if (SB_VERSION(&g_sb) == REISERFS_VERSION_2)
    {
      reiserfsck_paste_into_item (path, dir, DEH_SIZE + ROUND_UP (strlen ("..")) );
    }else{
      reiserfsck_paste_into_item (path, dir, DEH_SIZE + 2);
    }
}


static void insert_dot_and_dot_dot (struct path * path)
{
    struct item_head ih;
    char dir[EMPTY_DIR_SIZE];
    struct reiserfs_de_head * deh;

    copy_key (&ih.ih_key, &root_key);

    set_key_format (&ih, KEY_FORMAT_1);
    set_type_and_offset (KEY_FORMAT_1, &ih.ih_key, DOT_OFFSET, TYPE_DIRENTRY);
    
    deh = (struct reiserfs_de_head *)dir;

    if (SB_VERSION(&g_sb) == REISERFS_VERSION_2)
    {
	ih.ih_item_len = cpu_to_le16 (EMPTY_DIR_SIZE);
	deh[0].deh_location = cpu_to_le16 (ih_item_len (&ih) - ROUND_UP (strlen (".")));
	deh[1].deh_location = cpu_to_le16 (deh_location (&deh[0]) - ROUND_UP (strlen ("..")));
    }
    else
    {
	ih.ih_item_len = cpu_to_le16 (EMPTY_DIR_SIZE_V1);
	deh[0].deh_location = cpu_to_le16 (ih_item_len (&ih) - 1/*strlen (".")*/);
	deh[1].deh_location = cpu_to_le16 (deh_location (&deh[0]) - 2/*strlen ("..")*/);
    }

    set_entry_count (&ih, 2);
    mark_item_unreachable (&ih);

    deh[0].deh_offset = cpu_to_le32 (DOT_OFFSET);
    deh[0].deh_dir_id = cpu_to_le32 (root_key.k_dir_id);
    deh[0].deh_objectid = cpu_to_le32 (root_key.k_objectid);
    deh[0].deh_state = 0;
    set_bit (DEH_Visible, &(deh[0].deh_state));

    deh[1].deh_offset = cpu_to_le32 (DOT_DOT_OFFSET);
    deh[1].deh_dir_id = 0;
    deh[1].deh_objectid = cpu_to_le32 (root_key.k_dir_id);
    deh[1].deh_state = 0;
    set_bit (DEH_Visible, &(deh[1].deh_state));
    dir[DEH_SIZE * 2] = '.';
    dir[DEH_SIZE * 2 + 1] = '.';
    dir[DEH_SIZE * 2 + 2] = '.';

    reiserfsck_insert_item (path, &ih, dir);
}


/* returns buffer, containing found directory item.*/
char * get_next_directory_item (struct path * path, struct key * key, struct key * parent, struct item_head * ih)
{
    char * dir_item;
    struct key * rdkey;
    struct buffer_head * bh;
    struct reiserfs_de_head * deh;
    int i;
    int retval;
    int length;

    if ((retval = usearch_by_entry_key (&g_sb, key, path)) != POSITION_FOUND) {
	if (get_offset (key) != DOT_OFFSET)
	    die ("get_next_directory_item: entry not found");

	/* first directory item not found */
	if (is_rootdir_entry_key (key)) {
	    /* add "." and ".." to the root directory */
            if (opt_fsck_mode == FSCK_SEMANTIC)
                reiserfs_panic(&g_sb, "get_next_directory_item: root has no \".\" entry %k\n", key);

	    if (retval == POSITION_NOT_FOUND)
		paste_dot_and_dot_dot (path);
	    else if (retval == DIRECTORY_NOT_FOUND)
		insert_dot_and_dot_dot (path);
	    else
		die ("get_next_directory_item: invalid return value");
	    usearch_by_entry_key (&g_sb, key, path);
	} else {
	    /* it is ok for directories but the root one that "." is not found */
	    pathrelse (path);
	    return 0;
	}
    }
    /* leaf containing directory item */
    bh = PATH_PLAST_BUFFER (path);

    memcpy (ih, PATH_PITEM_HEAD (path), IH_SIZE);

    /* make sure, that ".." exists as well */
    if (get_offset (key) == DOT_OFFSET) {
	if (ih_entry_count (ih) < 2) {
	    pathrelse (path);
	    return 0;
	}
	deh = B_I_DEH (bh, ih) + 1;
	if (SB_VERSION(&g_sb) == REISERFS_VERSION_2)
	{
	  length = ROUND_UP (strlen (".."));
	}else{
	  length = strlen ("..");
	}
	if (entry_length (ih, deh, 1) != length || memcmp ("..", name_in_entry (deh, 1), 2)) {
	    reiserfs_warning ("get_next_directory_item: \"..\" not found\n");
	    pathrelse (path);
	    return 0;
	}
    }

    deh = B_I_DEH (bh, ih);

    /* mark hidden entries as visible, reset ".." correctly */
    for (i = 0; i < ih_entry_count (ih); i ++, deh ++) {
	if (de_hidden (deh))
	{
            if (opt_fsck_mode == FSCK_SEMANTIC)
                reiserfs_panic(&g_sb, "get_next_directory_item: item %k has hidden entry %d\n", key, i);
	
	    if (opt_verbose)
		reiserfs_warning ("\nget_next_directory_item: hidden entry %d\n", i);

	    mark_de_visible (deh);
	    mark_buffer_dirty (bh, 0);
	}
	
	if (deh->deh_offset == DOT_OFFSET)
	{
	    if (comp_short_keys (&(deh->deh_dir_id), key) &&
		deh->deh_objectid != REISERFS_ROOT_PARENT_OBJECTID)
            {
                if (opt_fsck_mode == FSCK_SEMANTIC)
                    reiserfs_panic(&g_sb, "get_next_directory_item: wrong \".\" found %k\n", key);
		if (opt_verbose)
		    reiserfs_warning ("\nget_next_directory_item: \".\" fixed\n");
		deh->deh_dir_id = key->k_dir_id;
		deh->deh_objectid = key->k_objectid;
		mark_buffer_dirty (bh, 0);
	    }
	}
	
	if (deh->deh_offset == DOT_DOT_OFFSET)
	{
	    /* set ".." so that it points to the correct parent directory */
	    if (comp_short_keys (&(deh->deh_dir_id), parent) &&
		deh->deh_objectid != REISERFS_ROOT_PARENT_OBJECTID)
            {
                if (opt_fsck_mode == FSCK_SEMANTIC)
                    reiserfs_panic(&g_sb, "get_next_directory_item: wrong \"..\" found %k\n", key);
		if (opt_verbose)
		    reiserfs_warning ("\nget_next_directory_item: \"..\" fixed\n");
		deh->deh_dir_id = parent->k_dir_id;
		deh->deh_objectid = parent->k_objectid;
		mark_buffer_dirty (bh, 0);
	    }
	}
    }


    /* copy directory item to the temporary buffer */
    dir_item = getmem (ih_item_len (ih)); 
    memcpy (dir_item, B_I_PITEM (bh, ih), ih_item_len (ih));

    /* next item key */
    if (PATH_LAST_POSITION (path) == (B_NR_ITEMS (PATH_PLAST_BUFFER (path)) - 1) &&
	(rdkey = uget_rkey (path)))
	copy_key (key, rdkey);
    else {
	key->k_dir_id = 0;
	key->k_objectid = 0;
    }

    if (opt_fsck_mode != FSCK_SEMANTIC)
        mark_item_reachable (PATH_PITEM_HEAD (path), PATH_PLAST_BUFFER (path));
    return dir_item;
}


// get key of an object pointed by direntry and the key of the entry itself
static void get_object_key (struct reiserfs_de_head * deh, struct key * key, 
			    struct key * entry_key, struct item_head * ih)
{
    key->k_dir_id = deh->deh_dir_id;
    key->k_objectid = deh->deh_objectid;
    key->u.k_offset_v1.k_offset = SD_OFFSET;
    key->u.k_offset_v1.k_uniqueness = V1_SD_UNIQUENESS;

    entry_key->k_dir_id = ih->ih_key.k_dir_id;
    entry_key->k_objectid = ih->ih_key.k_objectid;
    entry_key->u.k_offset_v1.k_offset = deh->deh_offset;
    entry_key->u.k_offset_v1.k_uniqueness = DIRENTRY_UNIQUENESS;
}


static void reiserfsck_cut_entry (struct key * key)
{
    INITIALIZE_PATH (path);
    struct item_head * ih;

    if (usearch_by_entry_key (&g_sb, key, &path) != POSITION_FOUND || get_offset (key) == DOT_OFFSET)
	die ("reiserfsck_cut_entry: entry not found");

    ih = get_ih (&path);
    if (ih_entry_count (ih) == 1)
	reiserfsck_delete_item (&path);
    else {
	struct reiserfs_de_head * deh = B_I_DEH (get_bh (&path), ih) + path.pos_in_item;

	reiserfsck_cut_from_item (&path, -(DEH_SIZE + entry_length (ih, deh, path.pos_in_item)));
    }
}



/* check recursively the semantic tree. Returns 0 if entry points to
   good object, and -1 or -2 if this entry must be deleted (stat data
   not found or directory does have any items).  Hard links are not
   allowed, but if directory rename has been interrupted by the system
   crash, it is possible, that fsck will find two entries (not "..") 
   pointing to the same directory. In this case fsck keeps only the
   first one. */
#define OK 0
#define STAT_DATA_NOT_FOUND -1
#define DIRECTORY_HAS_NO_ITEMS -2

static __u32 stat_datas = 0;


// FIXME: hash can be detected wrong when two hash functions give the
// same value on the first name which gets here
static void detect_check_and_set_hash (struct super_block * s, char * name, int namelen,
				       __u32 deh_hash_value)
{
    int hash_code;
    
    if ((namelen == 1 && !strncmp (name, ".", 1)) ||
	(namelen == 2 && !strncmp (name, "..", 2)))
	return;

    if (deh_hash_value == GET_HASH_VALUE(keyed_hash (name, namelen))){
	hash_code = TEA_HASH;
    } else if (deh_hash_value == GET_HASH_VALUE(yura_hash (name, namelen))) {
	hash_code = YURA_HASH;
    } else if (deh_hash_value == GET_HASH_VALUE(r5_hash (name, namelen))) {
	hash_code = R5_HASH;
    } else {
	hash_code = TEA_HASH; // to avoid compiler's warning
	die ("check_semantic_tree: unknown hash is used");
    }

    if (!s->u.reiserfs_sb.s_hash_function) {
	// set hash function as it is not set yet
//	if (opt_fsck_mode == FSCK_SEMANTIC)
//	        die("detect_check_and_set_hash: hash is not set in superblock");
	
	s->u.reiserfs_sb.s_hash_function = code2function (hash_code);
	set_hash (s->u.reiserfs_sb.s_rs, hash_code);
    } else {
	// check whether other hashes were used
	if (hash_code != rs_hash (s->u.reiserfs_sb.s_rs))
	    printf ("check_semantic_tree: \"%s\" hash is used. Should be \"%s\" only\n", hash_name (hash_code), 
		    hash_name (rs_hash (s->u.reiserfs_sb.s_rs)));
    }
}


int check_semantic_tree (struct key * key, struct key * parent, int is_dot_dot)
{
    struct path path;
    void * sd;
    int version;
    __u32 nlink;
    __u32 entry_len = 0;

    if (!KEY_IS_STAT_DATA_KEY (key))
	die ("check_semantic_tree: key must be key of a stat data");

    /* look for stat data of an object */
    if (usearch_by_key (&g_sb, key, &path) == ITEM_NOT_FOUND) {
	if (is_rootdir_key (key) && opt_fsck_mode != FSCK_SEMANTIC) {
	    /* stat data of the root directory not found. Make it */
	    create_root_directory (&path);
	    usearch_by_key (&g_sb, key, &path);
	} else {
	    pathrelse (&path);
	    return STAT_DATA_NOT_FOUND;
	}
    }

    version = ih_key_format (get_ih (&path));
    sd = get_item(&path);

    if (version == KEY_FORMAT_2)
    {
	struct stat_data * sd_v2 = sd;

	if ( (sd_v2->sd_nlink == 0) && ( opt_fsck == 0 ) && (opt_fsck_mode != FSCK_SEMANTIC))
	    print_how_far (&stat_datas, get_event (STAT_DATA_ITEMS));
	
        if ((sd_v2->sd_nlink == 0) && (opt_fsck_mode == FSCK_SEMANTIC))
            reiserfs_panic(&g_sb, "check_semantic_tree: sd_nlink is 0 %k\n", &(get_ih(&path)->ih_key));
	
	if ( !S_ISDIR (sd_v2->sd_mode) )
	{
	    /* object is not a directory (regular, symlink, device file) */
	    check_regular_file (&path, sd);
	    pathrelse (&path);
	    return OK;
	}
	
	if (opt_fsck_mode != FSCK_SEMANTIC)
	{
            nlink = le32_to_cpu (sd_v2->sd_nlink) + 1;
            sd_v2->sd_nlink = cpu_to_le32 (nlink +
		((key->k_objectid == REISERFS_ROOT_OBJECTID && nlink == 1) ? 1 : 0));
	}
    }
    else
    {
	struct stat_data_v1 * sd_v1 = sd;
	
	if ((sd_v1->sd_nlink == 0) && ( opt_fsck == 0 ) && (opt_fsck_mode != FSCK_SEMANTIC))
	    print_how_far (&stat_datas, get_event (STAT_DATA_ITEMS));

        if ((sd_v1->sd_nlink == 0) && (opt_fsck_mode == FSCK_SEMANTIC))
            reiserfs_panic(&g_sb, "check_semantic_tree: sd_nlink is 0 %k\n", &(get_ih(&path)->ih_key));
	
	if (!S_ISDIR (sd_v1->sd_mode))
	{
	    /* object is not a directory (regular, symlink, device file) */
	    check_regular_file (&path, sd);
	    pathrelse (&path);
	    return OK;
	}
	
	if (opt_fsck_mode != FSCK_SEMANTIC)
	{
            nlink = le16_to_cpu (sd_v1->sd_nlink) + 1;
            sd_v1->sd_nlink = cpu_to_le16 (nlink +
                ((key->k_objectid == REISERFS_ROOT_OBJECTID && nlink == 1) ? 1 : 0));
        }
    }

    if (opt_fsck_mode != FSCK_SEMANTIC)
        mark_buffer_dirty (PATH_PLAST_BUFFER (&path), 0);

    /* object is directory */
    if (nlink == 1 || opt_fsck_mode == FSCK_SEMANTIC) {
	char * dir_item;
	struct item_head ih;
	struct key item_key, entry_key, object_key;
	__u64 dir_size = 0;
        __u32 blocks;

	/*print_how_far (&stat_datas, get_event (STAT_DATA_ITEMS));*/

	add_event (DIRECTORIES);
	copy_key (&item_key, key);
	item_key.u.k_offset_v1.k_offset = DOT_OFFSET;
	item_key.u.k_offset_v1.k_uniqueness = DIRENTRY_UNIQUENESS;
	pathrelse (&path);
	while ((dir_item = get_next_directory_item (&path, &item_key, parent, &ih)) != 0) {
	    /* dir_item is copy of the item in separately allocated memory */
	    int i;
	    int retval;
	    struct reiserfs_de_head * deh = (struct reiserfs_de_head *)dir_item + path.pos_in_item;

/*&&&&&&&&&&&&&&&*/
	    if (dir_size == 0) {
		if (deh->deh_offset != DOT_OFFSET || (deh + 1)->deh_offset != DOT_DOT_OFFSET)
		    die ("check_semantic_tree: Directory without \".\" or \"..\"");
	    }
/*&&&&&&&&&&&&&&&*/

	    for (i = path.pos_in_item; i < ih_entry_count (&ih); i ++, deh ++) {
		detect_check_and_set_hash (&g_sb, name_in_entry (deh, i), name_length (&ih, deh, i),
					   GET_HASH_VALUE(deh_offset (deh)));

		get_object_key (deh, &object_key, &entry_key, &ih);
		if (opt_fsck_mode != FSCK_SEMANTIC)
                    retval = check_semantic_tree (&object_key, key,
			        (deh->deh_offset == DOT_OFFSET || deh->deh_offset == DOT_DOT_OFFSET) ? 1 : 0);
		else
		    if (deh->deh_offset != DOT_OFFSET && deh->deh_offset != DOT_DOT_OFFSET)
		    {
		        add_path_key(&object_key);
	                print(name_in_entry(deh, i), name_length (get_ih(&path), deh, i));
                        retval = check_semantic_tree (&object_key, key, 0);
                        erase_print(name_length (&ih, deh, i));
		        del_path_key();
                    }else
                        retval = OK;

		if (retval != OK)
		{
		    if (opt_fsck_mode == FSCK_SEMANTIC)
                        reiserfs_panic(&g_sb, "check_semantic_tree: bad entry %d in %k\n", i, &ih.ih_key);

		    if (get_offset (&entry_key) == DOT_DOT_OFFSET && object_key.k_objectid == REISERFS_ROOT_PARENT_OBJECTID) {
			/* ".." of root directory can not be found */
			if (retval != STAT_DATA_NOT_FOUND)
			    die ("check_semantic_tree: stat data of parent directory of root directory found");
			dir_size += DEH_SIZE + ((version == KEY_FORMAT_2) ? ROUND_UP (strlen ("..")) : strlen (".."));
			continue;
		    }
		    add_event (DELETED_ENTRIES);
		    reiserfsck_cut_entry (&entry_key);
		} else {
		    /* OK */
                    dir_size += DEH_SIZE + entry_length (&ih, deh, i);
		}
	    }

	    freemem (dir_item);

	    if (not_of_one_file (&item_key, key)) {
		pathrelse (&path);
		break;
	    }
	    pathrelse (&path);
	}

	if (dir_size == 0)
	    return DIRECTORY_HAS_NO_ITEMS;

	if (usearch_by_key (&g_sb, key, &path) != ITEM_FOUND)
	    die ("check_semantic_tree: stat data not found");

/*        if (mark_id_as_used(PATH_PITEM_HEAD (&path)->ih_key.k_objectid))
            reiserfs_warning("check_semantic_tree: objectid is used already %k \n",
                        &(PATH_PITEM_HEAD (&path)->ih_key));*/
	
	if (opt_fsck_mode == FSCK_SEMANTIC)
	{
	    if (!is_objectid_used(&g_sb, PATH_PITEM_HEAD (&path)->ih_key.k_objectid))
	         die("check_semantic_tree: objectid %d is unused", PATH_PITEM_HEAD (&path)->ih_key.k_objectid);
	}else
            mark_objectid_as_used (&g_sb, PATH_PITEM_HEAD (&path)->ih_key.k_objectid);

        sd = get_item(&path);
        blocks = ceill((double)dir_size / 512);
        if (version == KEY_FORMAT_2)
        {
	    struct stat_data * sd_v2 = sd;

	    if (dir_size != le64_to_cpu (sd_v2->sd_size))
	    {
	        if (opt_fsck_mode == FSCK_SEMANTIC)
	            reiserfs_panic(&g_sb, "check_semantic_tree: bad dir size %k\n", &(get_ih(&path)->ih_key));
	
		add_event (FIXED_SIZE_DIRECTORIES);
		sd_v2->sd_size = cpu_to_le64 (dir_size);
	    }
	    if (blocks != le32_to_cpu (sd_v2->sd_blocks))
	    {
	        if (opt_fsck_mode == FSCK_SEMANTIC)
	            reiserfs_panic(&g_sb, "check_semantic_tree: bad blocks number %k\n", &(get_ih(&path)->ih_key));
	
	        sd_v2->sd_blocks = cpu_to_le32(blocks);
	    }
        }
	else
	{
	    struct stat_data_v1 * sd_v1 = sd;

	    if (dir_size != le32_to_cpu (sd_v1->sd_size))
	    {
	        if (opt_fsck_mode == FSCK_SEMANTIC)
	            reiserfs_panic(&g_sb, "check_semantic_tree: bad dir size %k\n", &(get_ih(&path)->ih_key));
		
		add_event (FIXED_SIZE_DIRECTORIES);
		sd_v1->sd_size = cpu_to_le32 (dir_size);
	    }
	    if (blocks != le32_to_cpu (sd_v1->u.sd_blocks))
	    {
	        if (opt_fsck_mode == FSCK_SEMANTIC)
	            reiserfs_panic(&g_sb, "check_semantic_tree: bad blocks number %k\n", &(get_ih(&path)->ih_key));
	
	        sd_v1->u.sd_blocks = cpu_to_le32(blocks);	
	    }
        }

	/* stat data of a directory is accessed */
	if (opt_fsck_mode != FSCK_SEMANTIC)
	    mark_item_reachable (PATH_PITEM_HEAD (&path), PATH_PLAST_BUFFER (&path));
    } else {
	/* we have accessed directory stat data not for the first time. we
	   can come here only from "." or "..". Other names must be removed
	   to avoid creation of hard links */
	if (opt_fsck_mode == FSCK_SEMANTIC)
            die ("check_semantic_tree: can not get here");
	
	if (!is_dot_dot) {
	    if (version == KEY_FORMAT_2)
	    {
		struct stat_data * sd_v2 = sd;

		nlink = le32_to_cpu (sd_v2->sd_nlink);
		sd_v2->sd_nlink = cpu_to_le32 (nlink - 1);
	    }
	    else
	    {
		struct stat_data_v1 * sd_v1 = sd;

		nlink = le16_to_cpu (sd_v1->sd_nlink);
		sd_v1->sd_nlink = cpu_to_le16 (nlink - 1);
	    }

	    if (opt_verbose)
		reiserfs_warning ("\ncheck_semantic_tree: more than one name (neither \".\" nor \"..\") of a directory. Removed\n");
	    pathrelse (&path);
	    return STAT_DATA_NOT_FOUND;
	}
    }
    pathrelse (&path);


    return OK;
}

struct key g_root_directory_key = {REISERFS_ROOT_PARENT_OBJECTID, REISERFS_ROOT_OBJECTID, {{0, 0},}};
struct key g_parent_root_directory_key = {0, REISERFS_ROOT_PARENT_OBJECTID, {{0, 0},}};

void semantic_pass (void)
{
    if (opt_stop_point == STOP_AFTER_PASS1 || opt_stop_point == STOP_AFTER_PASS2)
	return;

//    init_objectid_list();

    if ( opt_fsck == 0 )
	fprintf (stderr, "Pass 3 (semantic) - ");
    if (check_semantic_tree (&g_root_directory_key, &g_parent_root_directory_key, 0) != OK)
        die ("check_semantic_tree: bad root found");
    if ( opt_fsck == 0 )
	printf ("\n");

//    free_objectid_maps();
}

void semantic_check (void)
{
    printf ("Checking Semantic tree...\n");

//    init_objectid_list();

    if (check_semantic_tree (&g_root_directory_key, &g_parent_root_directory_key, 0) != OK)
        die ("check_semantic_tree: bad root found");

//    free_objectid_maps();

    printf ("\n Ok\n");

}






























































