/*
 * Copyright 1996, 1997, 2000 Hans Reiser
 */
#include "fsck.h"
#include <math.h>

#if 0 /* original function */
//
// FIXME: we are going to turn reuse of objectids back
//
__u32 get_unused_objectid (struct super_block * s)
{
    __u32 * map, objectid;

    map = (__u32 *)((char *)SB_DISK_SUPER_BLOCK (s) + (sb_size(s)));
    objectid = map[1];
    if (objectid == TYPE_INDIRECT)
	die ("get_unused_objectid: no more objectids");

    map[1] ++;
    return objectid;
}
#endif

/* stolen from reiserfs/objectid.c  --clm */
__u32 get_unused_objectid (struct super_block * s)
{
    __u32 unused_objectid;
    struct reiserfs_super_block * rs;
    __u32 * objectid_map;


    rs = s->u.reiserfs_sb.s_rs;

    /* The objectid map follows the superblock. */
    objectid_map = (__u32 *)((char *)rs + (sb_size(s)));
 
    unused_objectid = objectid_map[1]; /* commented more below */
    if (unused_objectid == TYPE_INDIRECT) {
	printf ("REISERFS: get_objectid: no more object ids\n");
	return 0;
    }

    /* This incrementation allocates the first unused objectid. That is to say, 
       the first entry on the objectid map is the first unused objectid, and 
       by incrementing it we use it.  See below where we check to see if we 
       eliminated a sequence of unused objectids.... */

    objectid_map[1] ++;

    /* Now we check to see if we eliminated the last remaining member of
       the first even sequence (and can eliminate the sequence by
       eliminating its last objectid from oids), and can collapse the
       first two odd sequences into one sequence.  If so, then the net
       result is to eliminate a pair of objectids from oids.  We do this
       by shifting the entire map to the left. */
    if (rs_objectid_map_size (rs) > 2 && objectid_map[1] == objectid_map[2]) {
	memmove (objectid_map + 1, objectid_map + 3, (rs_objectid_map_size (rs) - 3) * sizeof(__u32));
	set_objectid_map_size (rs, rs_objectid_map_size (rs) - 2);
    }

    return unused_objectid;
}

int is_objectid_used (struct super_block * s, __u32 objectid)
{
    __u32 * objectid_map;
    int i = 0;

    objectid_map = (__u32 *)((char *)SB_DISK_SUPER_BLOCK (s) + (sb_size(s)));
  
    while (i < SB_OBJECTID_MAP_SIZE (s)) {
	if (objectid == objectid_map[i]) {
	    return 1;      /* objectid is used */
	}
	
	if (objectid > objectid_map[i] && objectid < objectid_map[i+1]) {
	    return 1;	/* objectid is used */
	}
	
	if (objectid < objectid_map[i])
	    break;

	i += 2;
    }
    
    /* objectid is free */
    return 0;
}


/* we mark objectid as used. Additionally, some unused objectids can
   become used. It is ok. What is unacceptable, it is when used
   objectids are marked as unused */
void mark_objectid_as_used (struct super_block * s, __u32 objectid)
{
    int i;
    __u32 * objectid_map;
    
    if (is_objectid_used (s, objectid) == 1) {
	/*print_objectid_map (s);*/
	/*printf ("mark_objectid_as_used: objectid %lu is used", objectid);*/
	return;
    }

    objectid_map = (__u32 *)((char *)SB_DISK_SUPER_BLOCK (s) + (sb_size(s)));

    for (i = 0; i < SB_OBJECTID_MAP_SIZE (s); i += 2) {
	if (objectid >= le32_to_cpu (objectid_map [i]) && 
	    objectid < le32_to_cpu (objectid_map [i + 1]))
	    /* it is used */
	    return;

	if (objectid + 1 == le32_to_cpu (objectid_map[i])) {
	    /* size of objectid map is the same */
	    objectid_map[i] = cpu_to_le32 (objectid);
	    return;
	}

	if (objectid == le32_to_cpu (objectid_map[i + 1])) {
	    /* size of objectid map is decreased */
	    objectid_map[i + 1] = cpu_to_le32 (le32_to_cpu (objectid_map[i + 1]) + 1);
	    if (i + 2 < SB_OBJECTID_MAP_SIZE (s)) {
		if (objectid_map[i + 1] == objectid_map[i + 2]) {
		    memmove (objectid_map + i + 1, objectid_map + i + 1 + 2, 
			     (SB_OBJECTID_MAP_SIZE (s) - (i + 2 + 2 - 1)) * sizeof (__u32));
		    
		    set_objectid_map_size (s->u.reiserfs_sb.s_rs, SB_OBJECTID_MAP_SIZE (s) - 2);
		    //SB_DISK_SUPER_BLOCK (s)->s_oid_cursize -= 2;
		}
	    }
	    return;
	}
	
	if (objectid < le32_to_cpu (objectid_map[i])) {
	    /* size of objectid map must be increased */
	    if (SB_OBJECTID_MAP_SIZE (s) == SB_OBJECTID_MAP_MAXSIZE (s)) {
		/* here all objectids between objectid and objectid_map[i] get used */
		objectid_map[i] = cpu_to_le32 (objectid);
		return;
	    } else {
		memmove (objectid_map + i + 2, objectid_map + i, (SB_OBJECTID_MAP_SIZE (s) - i) * sizeof (__u32));
		set_objectid_map_size (s->u.reiserfs_sb.s_rs, SB_OBJECTID_MAP_SIZE (s) + 2);
	    }
	    
	    objectid_map[i] = cpu_to_le32 (objectid);
	    objectid_map[i+1] = cpu_to_le32 (objectid + 1);
	    return;
	}
	
    }
    
    /* write out of current objectid map, if we have space */
    if (i < SB_OBJECTID_MAP_MAXSIZE (s)) {
	objectid_map[i] = cpu_to_le32 (objectid);
	objectid_map[i + 1] = cpu_to_le32 (objectid + 1);
	set_objectid_map_size (s->u.reiserfs_sb.s_rs, SB_OBJECTID_MAP_SIZE (s) + 2);
    } else if (i == SB_OBJECTID_MAP_MAXSIZE (s)) { //SB_DISK_SUPER_BLOCK (&g_sb)->s_oid_maxsize) {
	objectid_map[i - 1] = cpu_to_le32 (objectid + 1);
    } else
	die ("mark_objectid_as_used: objectid map corrupted");
    
    return;
}

#define MAP_SIZE   1024 /* must be n * 2 * sizeof(__u32) */

struct{
    __u32 * m_begin;
    __u32   m_used;  // used places in m_count * MAP_SIZE
    __u32   m_count; // count of MAPs
}id_map;

void init_id_map()
{
    id_map.m_begin = NULL;
    id_map.m_used = 0;
    id_map.m_count = 0;
    mark_id_as_used(1);
}

void grow_id_map()
{
    id_map.m_begin = expandmem(id_map.m_begin, id_map.m_count * MAP_SIZE, MAP_SIZE);
    id_map.m_count++;
}

void check_map_size()
{
    if (id_map.m_begin == NULL)
        if (id_map.m_used != 0 || id_map.m_count != 0)
            die("check_map_size: corrupted map structure");

    if (id_map.m_begin != NULL)
        if (id_map.m_count == 0)
            die("check_map_size: corrupted map structure");

    if (id_map.m_used > id_map.m_count * MAP_SIZE)
        die("check_map_size: corrupted map structure");

    if (id_map.m_used * sizeof(__u32) <= (id_map.m_count - 1) * MAP_SIZE)
    {
        if (id_map.m_count != 1)
            id_map.m_begin = expandmem(id_map.m_begin, id_map.m_count * MAP_SIZE, -MAP_SIZE);
        else
            freemem(id_map.m_begin);
        id_map.m_count--;
    }
}

void free_id_map()
{
    id_map.m_used = 0;
    while(id_map.m_count)
        check_map_size();
}

//FIXME: pos must be __u32
int check_id_search_result(int res, __u32 pos, __u32 id)
{
    if (res != ITEM_FOUND && res != ITEM_NOT_FOUND)
        die("is_id_used: get wrong result from ubin_search (%d)", res);

    if (res == ITEM_FOUND && *(id_map.m_begin + pos) != id)
        die("is_id_used: wrong id found %u %u", id, *(id_map.m_begin + pos));

    if (res == ITEM_NOT_FOUND)
    {
//        if (pos == 0)
//            die("is_id_used: id (%u) less than first id in the map %u", id, *(id_map.m_begin));
//        if (pos < 0)
//            die("is_id_used: wrong search result %d", pos);
        if (pos > id_map.m_used)
            die("is_id_used: get bad position (%u), used %u", pos, id_map.m_used);
        if (pos >= 0 && pos <= id_map.m_used && *(id_map.m_begin + pos - 1) >= id)
            die("is_id_used: previous id (%u) not less than (%u)", *(id_map.m_begin + pos - 1), id);
        if (pos >= 0 && pos < id_map.m_used && *(id_map.m_begin + pos) < id)
            die("is_id_used: found id (%u) not much than (%u)", *(id_map.m_begin + pos), id);
    }
}

int is_id_used (__u32 id, __u32 * pos)
{
    int i = 0;
    int res;
    __u32 * m_curr;

    *pos = 0;
    if (id_map.m_begin != NULL)
    {
        res = ubin_search_id(&id, id_map.m_begin, id_map.m_used, pos);
        if (opt_verbose)
            check_id_search_result(res, *pos, id);
    }else
        return 0;

        //pos is either a needed id or a right bound
    if (id_map.m_begin == NULL)
        return 0;

    if (res == ITEM_NOT_FOUND){
        if (fmod(*pos, 2) == 0)
            return 0;
        else
            return 1;
    }else if (res == ITEM_FOUND){
        if (fmod(*pos, 2) == 0)
            return 1;
        else
            return 0;
    }
    die("is_id_used: should not get here");
}

/* returns 1 if error occurs, 0 otherwise */
int mark_id_as_used (__u32 id)
{
    __u32 pos;
    __u32 * objectid_map;

    if (is_id_used (id, &pos) == 1){
        /*print_objectid_map (s);*/
        return 1;
//        die("mark_id_as_used: id is used already");
    }

    if (fmod(pos, 2) == 0){
        /* ITEM_NOT_FOUND */
        if (id_map.m_begin == NULL)
            grow_id_map();

        if ( id + 1 == le32_to_cpu (id_map.m_begin[pos]) ){
	    id_map.m_begin[pos] = cpu_to_le32 (id);
	    return 0;
	}
        if (id_map.m_count*MAP_SIZE == id_map.m_used * sizeof(__u32))
            grow_id_map();

        if (id_map.m_used - pos > 0)
            memmove (id_map.m_begin + pos + 2, id_map.m_begin + pos, (id_map.m_used - pos) * sizeof (__u32));

        id_map.m_used += 2;
        id_map.m_begin[pos] = cpu_to_le32 (id);
        id_map.m_begin[pos+1] = cpu_to_le32 (id + 1);
        return 0;
    }else if (fmod(pos, 2) == 1){
        /* ITEM_FOUND */
        id_map.m_begin[pos] = cpu_to_le32(id + 1);
        if (id_map.m_begin[pos + 1] == id_map.m_begin[pos]) {
            memmove (id_map.m_begin + pos, id_map.m_begin + pos + 2, (id_map.m_used - pos - 2) * sizeof (__u32));
            id_map.m_used -=2;
            check_map_size();
        }
        return 0;
    }
    die("mark_is_as_used: should not get here");
}

void print_objectid_list ()
{
    int i;
    printf ("\n Map all %d, used:%d", id_map.m_count * MAP_SIZE, id_map.m_used);

    for (i = 0; i < id_map.m_used; i += 2)
	    printf ("[%u-%u]", id_map.m_begin[i], id_map.m_begin[i + 1] - 1);
}

/*
struct objectid_list
{
    struct id_map * m_first;
    __u32  m_number;
    int    last_map_size;
    int    status;

    struct curr_elem
    {
        int pos;
        struct id_map * map;
    } curr;

} map_list;

struct id_map
{
    __u32 * address;
    struct id_map * prev;
    struct id_map * next;
};

void init_objectid_list()
{
    map_list.m_first = NULL;
    map_list.curr.map = NULL;
    map_list.curr.pos = 0;
//    map_list.m_last = NULL;
    map_list.m_number = 0;
    map_list.status = 0;
    map_list.last_map_size = -1;

// mark as used root_parent objectid
    mark_id_as_used(1);
}

void grow_objectid_maps()
{   // get new 4096 bytes for objectid list
    struct id_map * new_map;

    new_map = getmem(sizeof(struct id_map));
    new_map->address = getmem(MAP_SIZE * sizeof (__u32));
    map_list.m_number++;

    if (map_list.m_first == NULL)
    {
        map_list.m_first = new_map;
        new_map->prev = new_map;
        new_map->next = new_map;
    }
    else
    {
        new_map->prev = map_list.m_first->prev;
        new_map->next = map_list.m_first;
        map_list.m_first->prev->next = new_map;
        map_list.m_first->prev = new_map;
    }
}

void cut_objectid_maps()
{   // free 4096 bytes from objectid list
    struct id_map * to_cut;

    if (map_list.m_first == NULL)
        die("can not delete a map from the empty list");

    to_cut = map_list.m_first->prev;
    if (map_list.m_first == to_cut)
    {
        map_list.m_first = NULL;
        map_list.curr.map = NULL;
        map_list.curr.pos = 0;
    }else if (map_list.curr.map == to_cut)
    {
        map_list.curr.map = map_list.curr.map->prev;
        map_list.curr.pos = MAP_SIZE - 1;
    }

    to_cut->prev->next = to_cut->next;
    to_cut->next->prev = to_cut->prev;

    freemem(to_cut->address);
    freemem(to_cut);

    map_list.m_number--;
}

void free_objectid_maps()
{
    while(map_list.m_first != NULL)
        cut_objectid_maps();
}

__u32 get_previous_id()
{
    if (map_list.curr.map == NULL)
        return 0;

    if (map_list.curr.pos > MAP_SIZE || map_list.curr.pos < 0)
        die("cannot get objectid on (%d) position", map_list.curr.pos - 1);

    if (map_list.curr.pos == 0)
    {
        if (map_list.curr.map == map_list.m_first)
            return 0;
        else
            return map_list.curr.map->prev->address[MAP_SIZE-1];
    }

    return map_list.curr.map->address[map_list.curr.pos - 1];
}

__u32 get_id()
{
    if (map_list.curr.map == NULL)
        return 0;

    if (map_list.curr.pos >= MAP_SIZE || map_list.curr.pos < 0)
        die("cannot get objectid on (%d) position", map_list.curr.pos);

    return map_list.curr.map->address[map_list.curr.pos];
}

__u32 get_next_id()
{
    if (map_list.curr.map == NULL)
        return 0;

    if (map_list.curr.pos >= MAP_SIZE || map_list.curr.pos < 0)
        die("cannot get objectid on (%d) position", map_list.curr.pos + 1);

    if (map_list.curr.pos == MAP_SIZE - 1)
    {
        if (map_list.curr.map == map_list.m_first->prev)
            return 0;
        else
            return map_list.curr.map->next->address[0];
    }

    return map_list.curr.map->address[map_list.curr.pos + 1];
}


int is_id_used(__u32 id)
{
    __u32 n_map, n_elem;

    if (map_list.m_first == NULL)
        return -1;

    //FIXME: we should find the map first and then the position

    for (   n_map = 0, map_list.curr.map = map_list.m_first;
            n_map < map_list.m_number;
            n_map++,   map_list.curr.map = map_list.curr.map->next)
    {
        for (map_list.curr.pos = 0; map_list.curr.pos < MAP_SIZE; map_list.curr.pos += 2)
        {
            if (id < get_id())
                return 0;       // id is unused and located between prev_id and id
            if (id >= get_id() && id < get_next_id())
                return 1;       // id is used already
            if (get_id() == 0)
                return -1;      // all used id are > then this one
	}
    }

    map_list.curr.map = map_list.m_first->prev;
    return -1;                  // all used id are > then this one
}

void set_id( int pos, __u32 id)
{// set id to (curr.map, pos)
    struct id_map * cur_map = map_list.curr.map;

    if (map_list.m_first == NULL)
    {
        grow_objectid_maps();
        map_list.curr.map = map_list.m_first;
        map_list.curr.pos = 0;
        map_list.last_map_size = 0;
        cur_map = map_list.m_first;
    }

    while (pos < 0)
    {
        if (cur_map == map_list.m_first)
            die("set_pos: wrong position");
        cur_map = cur_map->prev;
        pos += MAP_SIZE;
    }

    while (pos >= MAP_SIZE)
    {
        if (cur_map == map_list.m_first->prev)
        {
            grow_objectid_maps();
            map_list.last_map_size = 0;
        }
        cur_map = cur_map->next;
        pos -= MAP_SIZE;
    }

    cur_map->address[pos] = id;
//    if (cur_map == map_list.m_first->prev &&
//        map_list.last_map_size < pos + 1)
//        map_list.last_map_size = pos + 1;
}

void move_objectid_list(int inc)
{   // for last_map_size +- 2 only

    struct id_map * m_fir, * m_sec;
    int p_fir, p_sec;

    // DECREASE: move from curr.pos + 1(p_fir) to curr.pos - 1(p_sec)
    // INCREASE: move from curr.pos(p_fir) to curr.pos + 2(p_sec)
    // set first and second elements
    m_fir = m_sec = map_list.curr.map;
    if (inc == DECREASE)
    {
        p_sec = map_list.curr.pos + 1;
        if (map_list.curr.pos == 0)
        {
            if (map_list.curr.map == map_list.m_first)
                die("can not mark id (0) as used");
            else
            {
                m_fir = m_fir->prev;
                p_fir = MAP_SIZE - 1;
            }
        }else
            p_fir = map_list.curr.pos - 1;
    }else{
        p_fir = map_list.curr.pos;
        if (map_list.curr.pos == MAP_SIZE - 2)
        {
            if (map_list.curr.map == map_list.m_first->prev)
            {
                grow_objectid_maps();
                map_list.last_map_size = 0;
            }
            m_sec = m_sec->next;
            p_sec = 0;
        }else
            p_sec = map_list.curr.pos + 2;
    }


    if (inc == INCREASE)
    {
        map_list.last_map_size += 2;
        if (map_list.last_map_size > MAP_SIZE)
        {
            grow_objectid_maps();
            map_list.last_map_size -= MAP_SIZE;
        }
    }

    if (p_fir < MAP_SIZE - 2)
    {
        if (inc == INCREASE)
            memmove (m_sec->address + p_sec, m_fir->address + p_fir, (MAP_SIZE - p_sec) * sizeof (__u32));
        else
            memmove (m_fir->address + p_fir, m_sec->address + p_sec, (MAP_SIZE - p_sec) * sizeof (__u32));

        p_fir = MAP_SIZE - 2;
        p_sec = 0;
        m_sec = m_sec->next;
    }

    while(m_fir != map_list.m_first && m_sec != map_list.m_first)
    {
        if (inc == INCREASE)
            memmove (m_sec->address + p_sec, m_fir->address + p_fir, 2 * sizeof (__u32));
        else
            memmove (m_fir->address + p_fir, m_sec->address + p_sec, 2 * sizeof (__u32));
        p_fir = 0;
        p_sec = 2;
        m_fir = m_fir->next;

        if (inc == INCREASE)
            memmove (m_sec->address + p_sec, m_fir->address + p_fir, (MAP_SIZE - 2) * sizeof (__u32));
        else
            memmove (m_fir->address + p_fir, m_sec->address + p_sec, (MAP_SIZE - 2) * sizeof (__u32));

        p_fir = MAP_SIZE - 2;
        p_sec = 0;
        m_sec = m_sec->next;
    }

    if (inc == DECREASE)
        map_list.last_map_size -= 2;

    if (map_list.last_map_size <= 0)
    {
        cut_objectid_maps();
        map_list.last_map_size += MAP_SIZE;
    }
}

// it returns if id was used
int mark_id_as_used (__u32 id)
{
    int pos;
    int is_used = is_id_used (id);

    if (id == 0)
        die ("cannot mark as used root dir_id");

    switch(is_used)
    {
    case 1:
//	reiserfs_warning("mark_id_as_used: objectid %lu is used", id);
	return 1;
    case -1:
        //add id at the end
        // 1. curr == {?, ?} and m_first == null
        if (map_list.m_first == NULL)
        {
            set_id(0, id);
            set_id(1, id + 1);
            map_list.last_map_size += 2;
            return 0;
        }

        // 2. curr == existent position
        // 3. curr == {MAP_SIZE, "last"}
        if (id == get_previous_id())
        {
            set_id(map_list.curr.pos - 1, id + 1);
            return 0;
        }
        set_id(map_list.curr.pos, id);
        set_id(map_list.curr.pos + 1, id + 1);
        map_list.last_map_size += 2;
        return 0;
    case 0:
	if ( id == get_previous_id() )
	{
            set_id(map_list.curr.pos - 1, id + 1);
            if (get_previous_id() == get_id())
                move_objectid_list(DECREASE);
	    return 0;
	}
	
	if ( id + 1 == get_id() )
	{
	    set_id (map_list.curr.pos, id);
	    return 0;
	}

	// get_previous_id() < id < get_id() - 1
    //       size of objectid map must be increased
	
        move_objectid_list(INCREASE);
        set_id(map_list.curr.pos, id);
        set_id(map_list.curr.pos + 1, id + 1);
	//memmove (objectid_map + i + 2, objectid_map + i, (SB_OBJECTID_MAP_SIZE (s) - i) * sizeof (__u32));
	return 0;
    }
    die("should't get here");
}
*/
/*
   if we have to delete items and put them into the tree again it returns 0
   and it returns 1 if we can just change id to new id without moving items
   of the file.
*/
/*
__u32 get_free_id(__u32 begin, __u32 end, int * to_move)
{
    *to_move = 0;

    if (!is_id_used(begin))
        return begin;
    if (get_next_id() > end)
        *to_move = 1;
    return get_next_id();
}
*/




